# Author: Rishabh Sharma <rishabh.sharma.gunner@gmail.com>
# This module was developed under funding provided by
# Google Summer of Code 2014
import csv
import socket
from datetime import datetime
from urllib.error import URLError
from urllib.request import urlopen, urlretrieve

from dateutil.rrule import MONTHLY, rrule

from astropy.time import Time

from sunpy.extern.parse import parse
from sunpy.net.dataretriever import GenericClient, QueryResponse
from sunpy.time import TimeRange, parse_time

__all__ = ['RHESSIClient']

data_servers = ('https://hesperia.gsfc.nasa.gov/hessidata/',
                'http://hessi.ssl.berkeley.edu/hessidata/',
                'http://soleil.i4ds.ch/hessidata/')

lc_linecolors = ('black', 'pink', 'green', 'blue', 'brown', 'red',
                 'navy', 'orange', 'green')


def parse_observing_summary_dbase_file(filename):
    """
    Parse the RHESSI observing summary database file.
    This file lists the name of observing summary files
    for specific time ranges along with other info.

    Parameters
    ----------
    filename : `str`
        The filename of the obssumm dbase file.

    Returns
    -------
    `dict`
        Return a `dict` containing the parsed data in the dbase file.

    References
    ----------
    https://hesperia.gsfc.nasa.gov/ssw/hessi/doc/guides/hessi_data_access.htm#Observing%20Summary%20Data
    """
    # An example dbase file can be found at:
    # https://hesperia.gsfc.nasa.gov/hessidata/dbase/hsi_obssumm_filedb_200311.txt

    with open(filename) as fd:
        reader = csv.reader(fd, delimiter=" ", skipinitialspace=True)
        _ = next(reader)  # skip 'HESSI Filedb File:' row
        _ = next(reader)  # skip 'Created: ...' row
        _ = next(reader)  # skip 'Number of Files: ...' row
        column_names = next(reader)  # ['Filename', 'Orb_st', 'Orb_end',...]

        obssumm_filename = []
        orbit_start = []
        orbit_end = []
        start_time = []
        end_time = []
        status_flag = []
        number_of_packets = []

        for row in reader:
            obssumm_filename.append(row[0])
            orbit_start.append(int(row[1]))
            orbit_end.append(int(row[2]))
            start_time.append(Time.strptime(row[3], "%d-%b-%y"))  # skip time
            end_time.append(Time.strptime(row[5], "%d-%b-%y"))  # skip time
            status_flag.append(int(row[7]))
            number_of_packets.append(int(row[8]))

        return {
            column_names[0].lower(): obssumm_filename,
            column_names[1].lower(): orbit_start,
            column_names[2].lower(): orbit_end,
            column_names[3].lower(): start_time,
            column_names[4].lower(): end_time,
            column_names[5].lower(): status_flag,
            column_names[6].lower(): number_of_packets,
        }


def get_base_url():
    """
    Find the first mirror which is online
    """
    for server in data_servers:
        try:
            urlopen(server, timeout=1)
            return server
        except (URLError, socket.timeout):
            pass
    raise OSError(f'Unable to find an online HESSI server from {data_servers}')


class RHESSIClient(GenericClient):
    """
    Provides access to the RHESSI observing summary time series data.

    Uses this `archive <https://hesperia.gsfc.nasa.gov/hessidata/>`__ or its mirrors.

    Examples
    --------
    >>> from sunpy.net import Fido, attrs as a
    >>> results = Fido.search(a.Time("2016/1/1", "2016/1/2"),
    ...                       a.Instrument.rhessi, a.Physobs.summary_lightcurve)  #doctest: +REMOTE_DATA
    >>> results  #doctest: +REMOTE_DATA
    <sunpy.net.fido_factory.UnifiedResponse object at ...>
    Results from 1 Provider:
    <BLANKLINE>
    2 Results from the RHESSIClient:
    Source: https://hesperia.gsfc.nasa.gov/hessidata
    <BLANKLINE>
           Start Time               End Time        Instrument ... Source Provider
    ----------------------- ----------------------- ---------- ... ------ --------
    2016-01-01 00:00:00.000 2016-01-01 23:59:59.999     RHESSI ... RHESSI     NASA
    2016-01-02 00:00:00.000 2016-01-02 23:59:59.999     RHESSI ... RHESSI     NASA
    <BLANKLINE>
    <BLANKLINE>

    """
    pattern = '{}/catalog/hsi_obssumm_{year:4d}{month:2d}{day:2d}_{}'

    @property
    def info_url(self):
        return 'https://hesperia.gsfc.nasa.gov/hessidata'

    def get_observing_summary_filename(self, time_range):
        """
        Download the RHESSI observing summary data from one of the RHESSI
        servers, parses it, and returns the name of the observing summary files
        relevant for the time range.

        Parameters
        ----------
        time_range : str, `sunpy.time.TimeRange`
            A `sunpy.time.TimeRange` or time range compatible string.

        Returns
        -------
        out : list
            Returns the urls of the observation summary file

        Examples
        --------
        >>> from sunpy.net.dataretriever.sources.rhessi import RHESSIClient
        >>> RHESSIClient().get_observing_summary_filename(('2011/04/04', '2011/04/04'))   # doctest: +REMOTE_DATA
        ['https://.../hessidata/metadata/catalog/hsi_obssumm_20110404_058.fits']
        """
        dt = TimeRange(time_range)
        # remove time from dates
        dt = TimeRange(dt.start.strftime('%Y-%m-%d'), dt.end.strftime('%Y-%m-%d'))

        filenames = []

        diff_months = (dt.end.datetime.year - dt.start.datetime.year) * \
            12 + dt.end.datetime.month - dt.start.datetime.month
        first_month = datetime(dt.start.datetime.year, dt.start.datetime.month, 1)
        month_list = rrule(MONTHLY, dtstart=first_month, count=diff_months+1)

        # need to download and inspect the dbase file to determine the filename
        # for the observing summary data
        # the dbase files are monthly but contain the daily filenames
        for this_month in month_list:
            dbase_file_name, hdrs = self.get_observing_summary_dbase_file(this_month)
            dbase_dat = parse_observing_summary_dbase_file(dbase_file_name)
            this_month_obssumm_filenames = dbase_dat.get('filename')
            daily_filenames_dates = [datetime.strptime(
                d[0:20], 'hsi_obssumm_%Y%m%d') for d in this_month_obssumm_filenames]
            for i, this_date in enumerate(daily_filenames_dates):
                if dt.start <= this_date <= dt.end:
                    filenames.append(
                        get_base_url()+f'metadata/catalog/{this_month_obssumm_filenames[i]}s')

        return filenames

    @staticmethod
    def get_observing_summary_dbase_file(time):
        """
        Download the RHESSI observing summary database file for the time given.
        One file covers an entire month.  This file lists the name of observing
        summary files for specific times.

        Parameters
        ----------
        time : `str`, datetime

        Returns
        -------
        value : `tuple`
            Return a `tuple` (filename, headers) where filename is the local file
            name under which the object can be found, and headers is
            whatever the info() method of the object returned by urlopen.

        Examples
        --------
        >>> from sunpy.net.dataretriever.sources.rhessi import RHESSIClient
        >>> fname, headers = RHESSIClient.get_observing_summary_dbase_file('2011/04/04')  # doctest: +REMOTE_DATA

        References
        ----------
        | http://hesperia.gsfc.nasa.gov/ssw/hessi/doc/guides/hessi_data_access.htm#Observing Summary Data

        .. note::
            This API is currently limited to providing data from whole days only.

        """
        _time = parse_time(time)

        if _time < parse_time("2002/02/01"):
            raise ValueError("RHESSI summary files are not available before 2002-02-01")

        url = get_base_url() + f'dbase/{_time.strftime("hsi_obssumm_filedb_%Y%m.txt")}'
        return urlretrieve(url)

    def search(self, *args, **kwargs):
        _, pattern, matchdict = self.pre_search_hook(*args, **kwargs)
        timerange = TimeRange(matchdict['Start Time'], matchdict['End Time'])
        metalist = []
        for url in self.get_observing_summary_filename(timerange):
            exdict = parse(pattern, url).named
            exdict['url'] = url
            rowdict = super().post_search_hook(exdict, matchdict)
            metalist.append(rowdict)
        return QueryResponse(metalist, client=self)

    @classmethod
    def register_values(cls):
        from sunpy.net import attrs
        adict = {attrs.Instrument: [('RHESSI',
                                     'Reuven Ramaty High Energy Solar Spectroscopic Imager.')],
                 attrs.Physobs: [("summary_lightcurve", "A summary lightcurve.")],
                 attrs.Source: [('RHESSI', 'Reuven Ramaty High Energy Solar Spectroscopic Imager.')],
                 attrs.Provider: [('NASA', 'The National Aeronautics and Space Administration.')]}
        return adict



import json
import pathlib

import requests

import astropy.table
from astropy.time import Time

import sunpy.net.attrs as a
from sunpy.net.attr import and_
from sunpy.net.base_client import BaseClient, QueryResponseTable
from sunpy.net.cdaweb.attrs import Dataset
from .walker import walker

__all__ = ['CDAWEBClient']

_CDAS_BASEURL = 'https://cdaweb.gsfc.nasa.gov/WS/cdasr/1'
_CDAS_HEADERS = {'Accept': 'application/json'}
_CDAS_TIME_FMT = '%Y%m%dT%H%M%SZ'
_DATAVIEW = 'sp_phys'


class CDAWEBClient(BaseClient):
    """
    Provides access to query and download from the Coordinated Data Analysis Web (CDAWeb).

    Examples
    --------
    >>> from sunpy.net import Fido, attrs as a
    >>> print(a.cdaweb.Dataset)
    sunpy.net.cdaweb.attrs.Dataset
    <BLANKLINE>
    Dataset ID.
    <BLANKLINE>
    <BLANKLINE>
                        Attribute Name                    Client ...                                   Description
    ----------------------------------------------------- ------ ... --------------------------------------------------------------------------------
    a1_k0_mpa                                             CDAWEB ... LANL 2001 Magnetospheric Plasma Analyzer Key Parameters - Mike Henderson (LANL)
    a2_k0_mpa                                             CDAWEB ... LANL 2002 Magnetospheric Plasma Analyzer Key Parameters - Mike Henderson (LANL)
    ac_at_def                                             CDAWEB ... ACE Hourly RTN, GSE and J2000 GCI Attitude direction cosines - E. C. Stone (C...
    ...
    wi_strahl0_swe                                        CDAWEB ... Wind Solar Wind Experiment (SWE) strahl detector, ~1/2 sec solar wind electro...
    wi_sw_ion_dist_swe_faraday                            CDAWEB ... Faraday Cup orientation and charge flux - Keith Ogilvie (NASA GSFC)
    wild2_helio1day_position                              CDAWEB ... Position in heliocentric coordinates from SPDF Helioweb - Natalia Papitashvil...
    >>>
    >>> res = Fido.search(a.Time('2021/07/01', '2021/07/08'),
    ...                   a.cdaweb.Dataset('SOLO_L2_MAG-RTN-NORMAL-1-MINUTE')) #doctest: +REMOTE_DATA
    >>> res #doctest: +REMOTE_DATA
    <sunpy.net.fido_factory.UnifiedResponse object at ...>
    Results from 1 Provider:
    <BLANKLINE>
    7 Results from the CDAWEBClient:
    Source: https://cdaweb.gsfc.nasa.gov/index.html
    <BLANKLINE>
               Dataset                    Start time               End time
    ------------------------------- ----------------------- -----------------------
    SOLO_L2_MAG-RTN-NORMAL-1-MINUTE 2021-07-01 00:00:29.000 2021-07-01 23:59:30.000
    SOLO_L2_MAG-RTN-NORMAL-1-MINUTE 2021-07-02 00:00:29.000 2021-07-02 23:59:30.000
    SOLO_L2_MAG-RTN-NORMAL-1-MINUTE 2021-07-03 00:00:29.000 2021-07-03 23:59:30.000
    SOLO_L2_MAG-RTN-NORMAL-1-MINUTE 2021-07-04 00:00:29.000 2021-07-04 23:59:30.000
    SOLO_L2_MAG-RTN-NORMAL-1-MINUTE 2021-07-05 00:00:29.000 2021-07-05 23:59:30.000
    SOLO_L2_MAG-RTN-NORMAL-1-MINUTE 2021-07-06 00:00:29.000 2021-07-06 23:59:30.000
    SOLO_L2_MAG-RTN-NORMAL-1-MINUTE 2021-07-07 00:00:29.000 2021-07-07 23:59:30.000
    <BLANKLINE>
    <BLANKLINE>

    See Also
    --------
    sunpy.net.cdaweb.get_datasets : Find dataset IDs for a given observatory
    sunpy.net.cdaweb.get_observatory_groups : Get all observatories avaiable from CDAWeb
    """
    @property
    def info_url(self):
        return 'https://cdaweb.gsfc.nasa.gov/index.html'

    def search(self, *query, **kwargs):
        """
        Search for datasets provided by the Space Physics Data Facility.
        """
        query = and_(*query)
        queries = walker.create(query)

        results = []
        for query_parameters in queries:
            results.append(self._do_search(query_parameters))
        table = astropy.table.vstack(results)
        qrt = QueryResponseTable(table, client=self)
        qrt.hide_keys = ['URL']
        return qrt

    def _do_search(self, query):
        response = (self._get_remote_files(query['dataset'],
                                           query['begin_time'],
                                           query['end_time']))

        if 'FileDescription' not in response:
            raise RuntimeError

        stimes = [f['StartTime'] for f in response['FileDescription']]
        etimes = [f['EndTime'] for f in response['FileDescription']]
        urls = [f['Name'] for f in response['FileDescription']]

        return astropy.table.QTable(
            {'Dataset': [query['dataset']] * len(stimes),
             'Start time': Time.strptime(stimes, '%Y-%m-%dT%H:%M:%S.%fZ').iso,
             'End time': Time.strptime(etimes, '%Y-%m-%dT%H:%M:%S.%fZ').iso,
             'URL': urls})

    @staticmethod
    def _get_remote_files(dataset, start, end):
        # Get a list of files for a given dataset between start and end times
        start = start.strftime(_CDAS_TIME_FMT)
        end = end.strftime(_CDAS_TIME_FMT)
        url = '/'.join([
            _CDAS_BASEURL,
            'dataviews', _DATAVIEW,
            'datasets', dataset,
            'orig_data', f'{start},{end}'
        ])
        response = requests.get(url, headers=_CDAS_HEADERS)
        return response.json()

    def fetch(self, query_results, *, path, downloader, **kwargs):
        for row in query_results:
            fname = row['URL'].split('/')[-1]
            filepath = str(path).format(file=fname)
            # Manually cap max_splits at 3 to make CDAWeb happy
            max_splits = kwargs.get('max_splits', 3)
            max_splits = min(max_splits, 3)
            downloader.enqueue_file(row['URL'], filename=filepath, max_splits=max_splits)

    @classmethod
    def _can_handle_query(cls, *query):
        required = {Dataset, a.Time}
        query_attrs = {type(x) for x in query}
        return required == query_attrs

    @classmethod
    def _attrs_module(cls):
        return 'cdaweb', 'sunpy.net.cdaweb.attrs'

    @classmethod
    def register_values(cls):
        return cls.load_dataset_values()

    @staticmethod
    def load_dataset_values():
        from sunpy.net import attrs as a

        attrs_path = pathlib.Path(__file__).parent / 'data' / 'attrs.json'
        with open(attrs_path, 'r') as attrs_file:
            all_datasets = json.load(attrs_file)

        # Convert from dict to list of tuples
        all_datasets = [(id, desc) for id, desc in all_datasets.items()]
        return {a.cdaweb.Dataset: all_datasets}



"""
This module provies a non-comprehensive collection of solar physical constants.
"""
# TODO: Need better sources for some constants as well as error values.
import astropy.constants.astropyconst20 as astrocon
import astropy.units as u
from astropy.constants import Constant
from astropy.time import Time

__all__ = ['physical_constants']

physical_constants = {}

# references
gsfc_fact = "https://nssdc.gsfc.nasa.gov/planetary/factsheet/sunfact.html"
allen = "Allen's Astrophysical Quantities 4th Ed."
archinal = 'Archinal et al. 2018'
asplund = "Asplund et al. 2006"
fivian = "Fivian et al. 2008"
meeus = "Meeus 1998 Astronomical Algorithms 2nd Ed."

physical_constants['mass'] = astrocon.M_sun
physical_constants['radius'] = astrocon.R_sun
physical_constants['luminosity'] = astrocon.L_sun
physical_constants['mean distance'] = astrocon.au

# following needs error estimate if appropriate
physical_constants['perihelion distance'] = Constant('perihelion',
                                                     "Perihelion Distance",
                                                     1.471e11, 'm', 0,
                                                     allen, system='si')

# following needs error estimate if appropriate
physical_constants['aphelion distance'] = Constant('aphelion',
                                                   "Aphelion Distance",
                                                   1.521e11, 'm', 0,
                                                   allen, system='si')

physical_constants['age'] = Constant('age', "Age of the Sun",
                                     4.6e9, 'year', 0.1e9,
                                     allen, system='si')

# A solar flux  (sfu) is traditional measure of solar radio flux.
physical_constants['solar flux unit'] = Constant('sfu', "Solar flux unit",
                                                 1e-22, 'W m**-2 Hz**-1', 0,
                                                 allen, system='si')

# following needs error estimate if appropriate
physical_constants['visual magnitude'] = Constant('V',
                                                  "Apparent visual magnitude",
                                                  -26.75, '', 0,
                                                  allen, system='si')

# The Sun as viewed from Earth
physical_constants['average angular size'] = Constant('theta', "Semidiameter",
                                                      959.63, 'arcsec', 0,
                                                      allen, system='si')

# following needs error estimate if appropriate
physical_constants['surface area'] = Constant('A', "Surface area",
                                              6.087e18, 'm**2', 0,
                                              allen, system='si')

# following needs error estimate if appropriate
physical_constants['average density'] = Constant('rho', "Mean density",
                                                 1409, 'kg m**-3', 0,
                                                 allen, system='si')

# following needs error estimate if appropriate
physical_constants['surface gravity'] = Constant('g', "Surface gravity",
                                                 274, 'm s**-2', 0,
                                                 allen, system='si')

# following needs error estimate if appropriate
physical_constants['moment of inertia'] = Constant('I', "Moment of inertia",
                                                   5.7e54, 'kg m**-2', 0,
                                                   allen, system='si')

# following needs error estimate if appropriate
physical_constants['volume'] = Constant('V', "Volume",
                                        1.4122e27, 'm**3', 0,
                                        allen, system='si')

# following needs error estimate if appropriate
physical_constants['escape velocity'] = Constant('v',
                                                 "Escape velocity at surface",
                                                 6.177e5, 'm s**-1', 0,
                                                 allen, system='si')

physical_constants['oblateness'] = Constant('', "oblateness",
                                            8.01, 'marcsec', 0.14,
                                            fivian, system='si')

# the following constants need references and error estimates if appropriate
physical_constants['metallicity'] = Constant('Z', "Metallicity",
                                             0.0122, '', 0.0,
                                             asplund, system='si')

sunspot_cycle_exp = "Average duration of sunspot cycle"
physical_constants['sunspot cycle'] = Constant('', sunspot_cycle_exp,
                                               11.4, 'year', 0,
                                               "", system='si')

physical_constants['average intensity'] = Constant('I', "Mean Intensity",
                                                   2.009e7, 'W m**-2 sr**-1',
                                                   0,
                                                   "", system='si')

effect_temp_exp = "Effective black-body temperature"
physical_constants['effective temperature'] = Constant('T', effect_temp_exp,
                                                       5778.0, 'K', 0,
                                                       "", system='si')

mass_conv_exp = "Mass conversion rate"
physical_constants['mass conversion rate'] = Constant('dm/dt', mass_conv_exp,
                                                      4300e6, 'kg s**-1', 0,
                                                      "", system='si')

# following needs error estimate if appropriate
physical_constants['center density'] = Constant('rho_center', "Center density",
                                                1.622e5, 'kg m**-3', 0,
                                                gsfc_fact, system='si')

# following needs error estimate if appropriate
cent_temp_exp = "Center temperature"
physical_constants['center temperature'] = Constant('T_center', cent_temp_exp,
                                                    1.571e7, 'K', 0,
                                                    gsfc_fact, system='si')

# following needs error estimate if appropriate
abs_magn_exp = "Absolute magnitude"
physical_constants['absolute magnitude'] = Constant('M_abs', abs_magn_exp,
                                                    +4.83, '', 0,
                                                    gsfc_fact, system='si')
# following needs error estimate if appropriate
mean_energy_exp = "mean energy production"
physical_constants['mean energy production'] = Constant('', mean_energy_exp,
                                                        193.7e-6, 'J kg**-1',
                                                        0,
                                                        gsfc_fact, system='si')

# following needs error estimate if appropriate
physical_constants['ellipticity'] = Constant('', "ellipticity",
                                             5e-5, '', 0,
                                             gsfc_fact, system='si')

# following needs error estimate if appropriate
physical_constants['GM'] = Constant('mu', "standard gravitational parameter",
                                    132.712e6, 'km**3 s**-2', 0,
                                    gsfc_fact, system='si')

# longitude of the prime meridian (without light travel time to Earth
# and aberration effects) is 84.176 degrees eastward at J2000
physical_constants['W_0'] = Constant('W_0',
                                     'longitude of the prime meridian (epoch J2000.0)',
                                     84.176, 'deg',
                                     0, archinal,
                                     system='si')

# the definitional (fixed) rotation rate of the Sun relative to the stars (i.e., sidereal rotation rate)
physical_constants['sidereal rotation rate'] = Constant('', 'sidereal rotation rate',
                                                        14.1844, 'deg day**-1', 0,
                                                        archinal, system='si')

# time in Julian Days (TT) of the start of the first Carrington rotation
first_carrington_rotation = Time(2398167.4, format='jd', scale='tt')
physical_constants['first Carrington rotation (JD TT)'] = Constant('',
                                                                   'first Carrington '
                                                                   'rotation (JD TT)',
                                                                   first_carrington_rotation.tt.jd,
                                                                   'day', 0.1,
                                                                   meeus, system='si')

# length of the mean Carrington rotation as seen from Earth
# the rotation rate of the Sun appears to be slower by the rate at which the Earth orbits the Sun
period = 1 / (physical_constants['sidereal rotation rate'] / (360*u.deg) - 1 / u.yr)
physical_constants['mean synodic period'] = Constant('',
                                                     'mean synodic period',
                                                     period.to_value('day'), 'day', 0,
                                                     archinal, system='si')

# Sun's north pole is oriented RA=286.13 deg, Dec=63.87 deg in ICRS and HCRS
physical_constants['alpha_0'] \
    = Constant('alpha_0',
               'right ascension (RA) of the north pole (epoch J2000.0)',
               286.13, 'deg', 0, archinal, system='si')
physical_constants['delta_0'] \
    = Constant('delta_0',
               'declination of the north pole (epoch J2000.0)',
               63.87, 'deg', 0, archinal, system='si')



"""
==========================================================
Overplotting SRS active region locations on a magnetograms
==========================================================

How to find and plot the location of an active region on an HMI magnetogram.
"""
import matplotlib.pyplot as plt
import numpy as np

from astropy.coordinates import SkyCoord

import sunpy.coordinates
import sunpy.data.sample
import sunpy.map
from sunpy.io.special import srs

##############################################################################
# For this example, we will start with the sample data. We need an HMI file and
# use it to create a map, and the SRS table which contains a list of active
# regions. Both of these data can be downloaded with ``Fido``.

smap = sunpy.map.Map(sunpy.data.sample.HMI_LOS_IMAGE)
srs_table = srs.read_srs(sunpy.data.sample.SRS_TABLE)

##############################################################################
# We only need the rows which have 'ID' = 'I' or 'IA'.

srs_table = srs_table[np.logical_or(srs_table['ID'] == 'I', srs_table['ID'] == 'IA')]

##############################################################################
# Now we extract the latitudes, longitudes and the region numbers.
# :meth:`~astropy.visualization.wcsaxes.WCSAxes.plot_coord` will error on some
# versions of Astropy if a coordinate component contains a mask, but since
# none of the masks in these arrays here actually mask any elements, we simply
# remove the masks.

lats = srs_table['Latitude']
if hasattr(lats, 'mask'):
    lats = lats.unmasked
lngs = srs_table['Longitude']
if hasattr(lngs, 'mask'):
    lngs = lngs.unmasked
numbers = srs_table['Number']

##############################################################################
# Let's plot the results by defining coordinates for each location.

fig = plt.figure()
ax = fig.add_subplot(projection=smap)
# Passing vmin/vmax to ``plot`` does not work since
# a normalisation is set on the map. So we have to
# work around it like so:
smap.plot_settings["norm"].vmin = -150
smap.plot_settings["norm"].vmax = 150
smap.plot(axes=ax)
smap.draw_limb(axes=ax)
c = SkyCoord(lngs, lats, frame="heliographic_stonyhurst")
ax.plot_coord(c, 'o')
for num, lng, lat in zip(numbers, lngs.value, lats.value):
    ax.annotate(num, (lng, lat),
                xycoords=ax.get_transform('heliographic_stonyhurst'),
                color='red',
                fontweight='bold')

plt.show()



"""
This module provides a wrapper around the Helioviewer API.
"""
import json
import codecs
import urllib
from pathlib import Path
from collections import OrderedDict
from urllib.parse import urljoin
from urllib.request import urlopen

from astropy.utils.decorators import lazyproperty

import sunpy
import sunpy.util.parfive_helpers as parfive
from sunpy import log
from sunpy.time import parse_time
from sunpy.util.util import partial_key_match
from sunpy.util.xml import xml_to_dict

__all__ = ['HelioviewerClient']

HELIOVIEWER_API_URLS = [
    "https://api.helioviewer.org/",
    "https://helioviewer-api.ias.u-psud.fr/",
]


def check_connection(url):
    try:
        resp = urlopen(urljoin(url, "/v2/getDataSources/"))
        assert resp.getcode() == 200
        assert isinstance(json.loads(resp.read()), dict)
        return url
    except Exception as e:
        log.debug(f"Unable to connect to {url}:\n {e}")
        log.info(f"Connection to {url} failed. Retrying with different url.")
    return None


class HelioviewerClient:
    """Helioviewer.org Client"""

    def __init__(self, url=None):
        """
        Parameters
        ----------
        url : `str`
            Default URL points to the Helioviewer API.
        """
        if url is None:
            for url in HELIOVIEWER_API_URLS:
                if check_connection(url):
                    break

        if url is None:
            raise ValueError("No online helioviewer API can be found.")

        self._api = url

    @lazyproperty
    def data_sources(self):
        """
        We trawl through the return from ``getDataSources`` to create a clean
        dictionary for all available sourceIDs.

        Here is a list of all of them: https://api.helioviewer.org/docs/v2/#appendix_datasources
        """
        data_sources_dict = dict()
        datasources = self.get_data_sources()
        for name, observ in datasources.items():
            # TRACE only has measurements and is thus nested once
            if name == "TRACE":
                for instr, params in observ.items():
                    data_sources_dict[(name, None, None, instr)] = params['sourceId']
            else:
                for inst, detect in observ.items():
                    for wavelength, params in detect.items():
                        if 'sourceId' in params:
                            data_sources_dict[(name, inst, None, wavelength)] = params['sourceId']
                        else:
                            for wave, adict in params.items():
                                data_sources_dict[(name, inst, wavelength, wave)
                                                  ] = adict['sourceId']

        # Sort the output for printing purposes
        return OrderedDict(sorted(data_sources_dict.items(), key=lambda x: x[1]))

    def get_data_sources(self):
        """
        Return a hierarchical dictionary of the available datasources on helioviewer.org.

        This uses ``getDataSources`` from the Helioviewer API.

        Returns
        -------
        out : `dict`
            A dictionary containing meta-information for each data source that Helioviewer supports.
        """
        params = {"action": "getDataSources"}
        return self._get_json(params)

    def get_closest_image(self, date, observatory=None, instrument=None,
                          detector=None, measurement=None, source_id=None):
        """
        Finds the closest image available for the specified source and date.
        **This does not download any file.**

        This uses `getClosestImage <https://api.helioviewer.org/docs/v2/#OfficialClients>`_
        from the Helioviewer API.

        .. note::
            We can use ``observatory`` and ``measurement`` or ``instrument`` and ``measurement`` to
            get the value for source ID which can then be used to get required information.

        Parameters
        ----------
        date : `astropy.time.Time`, `str`
            A `~sunpy.time.parse_time` parsable string or `~astropy.time.Time`
            object for the desired date of the image
        observatory : `str`
            Observatory name
        instrument : `str`
            Instrument name
        detector : `str`
            Detector name
        measurement : `str`
            Measurement name
        source_id : `int`
            ID number for the required instrument/measurement.
            This can be used directly instead of using the previous parameters.

        Returns
        -------
        out : `dict`
            A dictionary containing meta-information for the closest image matched

        Examples
        --------
        >>> from sunpy.net import helioviewer
        >>> client = helioviewer.HelioviewerClient()  # doctest: +REMOTE_DATA
        >>> metadata = client.get_closest_image('2012/01/01', source_id=11)  # doctest: +REMOTE_DATA
        >>> print(metadata['date'])  # doctest: +REMOTE_DATA
        2012-01-01T00:00:07.000
        """
        if source_id is None:
            source_id = self._get_source_id((observatory, instrument, detector, measurement))

        params = {
            "action": "getClosestImage",
            "date": self._format_date(date),
            "sourceId": source_id
        }
        response = self._get_json(params)

        # Cast date string to Time
        response['date'] = parse_time(response['date'])

        return response

    def download_jp2(self, date, progress=True, observatory=None, instrument=None, detector=None,
                     measurement=None, source_id=None, directory=None, overwrite=False):
        """
        Downloads the JPEG 2000 that most closely matches the specified time and
        data source.

        This uses `getJP2Image <https://api.helioviewer.org/docs/v2/#JPEG2000>`_
        from the Helioviewer API.

        .. note::
            We can use ``observatory`` and ``measurement`` or ``instrument`` and ``measurement``
            to get the value for source ID which can then be used to get required information.

        Parameters
        ----------
        date : `astropy.time.Time`, `str`
            A string or `~astropy.time.Time` object for the desired date of the image
        progress : `bool`
            Defaults to True.
            If set to False, disables progress bars seen on terminal when
            downloading files.
        observatory : `str`
            Observatory name
        instrument : `str`
            Instrument name
        measurement : `str`
            Measurement name
        detector : `str`
            Detector name
        source_id : `int`
            ID number for the required instrument/measurement.
            This can be used directly instead of using the previous parameters.
        directory : `str`
            Directory to download JPEG 2000 image to.
        overwrite : `bool`
            Defaults to False.
            If set to True, will overwrite any files with the same name.

        Returns
        -------
        out : `str`
            Returns a filepath to the downloaded JPEG 2000 image.

        Examples
        --------
        >>> import sunpy.map
        >>> from sunpy.net import helioviewer
        >>> hv = helioviewer.HelioviewerClient()  # doctest: +REMOTE_DATA
        >>> filepath = hv.download_jp2('2012/07/03 14:30:00', observatory='SDO',
        ...                            instrument='HMI', detector=None, measurement='continuum')  # doctest: +REMOTE_DATA
        >>> filepath = hv.download_jp2('2012/07/03 14:30:00', observatory='SDO', measurement='continuum')  # doctest: +REMOTE_DATA
        >>> filepath = hv.download_jp2('2012/07/03 14:30:00', instrument='HMI', measurement='continuum')  # doctest: +REMOTE_DATA
        >>> aia = sunpy.map.Map(filepath)  # doctest: +REMOTE_DATA
        >>> aia.peek()  # doctest: +SKIP
        """
        if source_id is None:
            source_id = self._get_source_id((observatory, instrument, detector, measurement))

        params = {
            "action": "getJP2Image",
            "date": self._format_date(date),
            "sourceId": source_id,
        }

        return self._get_file(params, progress=progress, directory=directory, overwrite=overwrite)

    def get_jp2_header(self, date, observatory=None, instrument=None, detector=None, measurement=None, jp2_id=None):
        """
        Get the XML header embedded in a JPEG2000 image. Includes the FITS header as well as a section
        of Helioviewer-specific metadata.

        This uses `getJP2Header <https://api.helioviewer.org/docs/v2/#JPEG2000>`_
        from the Helioviewer API.

        .. note::
            We can use ``observatory`` and ``measurement`` or ``instrument`` and ``measurement``
            to get the value for source ID which can then be used to get required information.

        Parameters
        ----------
        date : `astropy.time.Time`, `str`
            A `~sunpy.time.parse_time` parsable string or `~astropy.time.Time`
            object for the desired date of the image
        observatory : `str`
            Observatory name
        instrument : `str`
            Instrument name
        measurement : `str`
            Measurement name
        detector : `str`
            Detector name
        jp2_id : `int`
            Unique JP2 image identifier.
            This can be used directly instead of using the previous parameters.

        Returns
        -------
        out : `dict`
            Returns a dictionary containing the header information of JPEG 2000 image.
            The returned dictionary may have either one or two keys: *fits* and *helioviewer*.

        Examples
        --------
        >>> from sunpy.net import helioviewer
        >>> hv = helioviewer.HelioviewerClient()  # doctest: +REMOTE_DATA
        >>> header = hv.get_jp2_header('2012/07/03', observatory='SDO',
        ...                            instrument='HMI', detector=None, measurement='continuum')  # doctest: +REMOTE_DATA
        >>> # The key 'fits' can be used to get the fits header information
        >>> fits_header = header['fits']  # doctest: +REMOTE_DATA
        >>> # The keys 'helioviewer' can be used to extract the helioviewer specific metadata.
        >>> helioviewer_meta_data = header['helioviewer']  # doctest: +REMOTE_DATA
        """
        if jp2_id is None:
            jp2_id = self.get_closest_image(
                date, observatory, instrument, detector, measurement)['id']

        params = {
            "action": "getJP2Header",
            "id": jp2_id,
        }

        responses = self._request(params)
        # Reads the output from HTTPResponse object and decodes it.
        responses = responses.read().decode('utf-8')
        return xml_to_dict(responses)['meta']

    def download_png(self, date, image_scale, layers, progress=True,
                     directory=None, overwrite=False, watermark=False,
                     events="", event_labels=False,
                     scale=False, scale_type="earth", scale_x=0, scale_y=0,
                     width=4096, height=4096, x0=0, y0=0,
                     x1=None, y1=None, x2=None, y2=None):
        """
        Downloads the PNG that most closely matches the specified time and
        data source.

        This function is different to `~sunpy.net.helioviewer.HelioviewerClient.download_jp2`.
        Here you get PNG images and return more complex images.

        For example you can return an image that has multiple layers composited together
        from different sources.
        Also mark solar features/events with an associated text label.
        The image can also be cropped to a smaller field of view.

        These parameters are not pre-validated before they are passed to Helioviewer API.
        See https://api.helioviewer.org/docs/v2/#appendix_coordinates for more information about
        what coordinates values you can pass into this function.

        This uses `takeScreenshot <https://api.helioviewer.org/docs/v2/#Screenshots>`_ from the Helioviewer API.

        .. note::

            Parameters ``x1``, ``y1``, ``x2`` and ``y2`` are set to `None`.
            If all 4 are set to values, then keywords: ``width``, ``height``, ``x0``, ``y0`` will be ignored.

        Parameters
        ----------
        date : `astropy.time.Time`, `str`
            A `~sunpy.time.parse_time` parsable string or `~astropy.time.Time` object
            for the desired date of the image
        image_scale : `float`
            The zoom scale of the image in arcseconds per pixel.
            For example, the scale of an AIA image is 0.6.
        layers : `str`
            Image datasource layer/layers to include in the screeshot.
            Each layer string is comma-separated with either:
            "[sourceId,visible,opacity]" or "[obs,inst,det,meas,visible,opacity]".
            Multiple layers are: "[layer1],[layer2],[layer3]".
        progress : `bool`, optional
            Defaults to True.
            If set to False, disables progress bars seen on terminal when
            downloading files.
        events : `str`, optional
            Defaults to an  empty string to indicate no feature/event annotations.
            List feature/event types and FRMs to use to annoate the image.
            Example could be "[AR,HMI_HARP;SPoCA,1]" or "[CH,all,1]"
        event_labels : `bool`, optional
            Defaults to False.
            Annotate each event marker with a text label.
        watermark : `bool`, optional
            Defaults to False.
            Overlay a watermark consisting of a Helioviewer logo and
            the datasource abbreviation(s) and timestamp(s) in the screenshot.
        directory : `str`, optional
            Directory to download JPEG 2000 image to.
        overwrite : bool, optional
            Defaults to False.
            If set to True, will overwrite any files with the same name.
        scale : `bool`, optional
            Defaults to False.
            Overlay an image scale indicator.
        scale_type : `str`, optional
            Defaults to Earth.
            What is the image scale indicator will be.
        scale_x : `int`, optional
            Defaults to 0 (i.e, in the middle)
            Horizontal offset of the image scale indicator in arcseconds with respect
            to the center of the Sun.
        scale_y : `int`, optional
            Defaults to 0 (i.e, in the middle)
            Vertical offset of the image scale indicator in arcseconds with respect
            to the center of the Sun.
        x0 : `float`, optional
            The horizontal offset from the center of the Sun.
        y0 : `float`, optional
            The vertical offset from the center of the Sun.
        width : `int`, optional
            Defaults to 4096.
            Width of the image in pixels.
        height : `int`, optional
            Defaults to 4096.
            Height of the image in pixels.
        x1 : `float`, optional
            Defaults to None
            The offset of the image's left boundary from the center
            of the sun, in arcseconds.
        y1 : `float`, optional
            Defaults to None
            The offset of the image's top boundary from the center
            of the sun, in arcseconds.
        x2 : `float`, optional
            Defaults to None
            The offset of the image's right boundary from the
            center of the sun, in arcseconds.
        y2 : `float`, optional
            Defaults to None
            The offset of the image's bottom boundary from the
            center of the sun, in arcseconds.

        Returns
        -------
        out : `str`
            Returns a filepath to the downloaded PNG image.

        Examples
        --------
        >>> from sunpy.net.helioviewer import HelioviewerClient
        >>> hv = HelioviewerClient()  # doctest: +REMOTE_DATA
        >>> file = hv.download_png('2012/07/16 10:08:00', 2.4,
        ...                        "[SDO,AIA,AIA,171,1,100]",
        ...                        x0=0, y0=0, width=1024, height=1024)   # doctest: +REMOTE_DATA
        >>> file = hv.download_png('2012/07/16 10:08:00', 4.8,
        ...                        "[SDO,AIA,AIA,171,1,100],[SOHO,LASCO,C2,white-light,1,100]",
        ...                        x1=-2800, x2=2800, y1=-2800, y2=2800)   # doctest: +REMOTE_DATA
        """
        params = {
            "action": "takeScreenshot",
            "date": self._format_date(date),
            "imageScale": image_scale,
            "layers": layers,
            "eventLabels": event_labels,
            "events": events,
            "watermark": watermark,
            "scale": scale,
            "scaleType": scale_type,
            "scaleX": scale_x,
            "scaleY": scale_y,
            # Returns the image which we do not want a user to change.
            "display": True
        }

        # We want to enforce that all values of x1, x2, y1, y2 are not None.
        # You can not use both scaling parameters so we try to exclude that here.
        if any(i is None for i in [x1, x2, y1, y2]):
            adict = {"x0": x0, "y0": y0,
                     "width": width, "height": height}
        else:
            adict = {"x1": x1, "x2": x2,
                     "y1": y1, "y2": y2}
        params.update(adict)

        return self._get_file(params, progress=progress, directory=directory, overwrite=overwrite)

    def is_online(self):
        """Returns True if Helioviewer is online and available."""
        try:
            self.get_data_sources()
        except urllib.error.URLError:
            return False

        return True

    def _get_json(self, params):
        """Returns a JSON result as a string."""
        reader = codecs.getreader("utf-8")
        response = self._request(params)
        return json.load(reader(response))

    def _get_file(self, params, progress=True, directory=None, overwrite=False):
        """Downloads a file and return the filepath to that file."""
        if directory is None:
            directory = Path(sunpy.config.get('downloads', 'download_dir'))
        else:
            directory = Path(directory).expanduser().absolute()

        downloader = parfive.Downloader(progress=progress, overwrite=overwrite)

        url = urllib.parse.urljoin(self._api,
                                   "?" + urllib.parse.urlencode(params))
        downloader.enqueue_file(url, path=directory)

        res = downloader.download()

        if len(res) == 1:
            return res[0]
        else:
            return res

    def _request(self, params):
        """
        Sends an API request and returns the result.

        Parameters
        ----------
        params : `dict`
            Parameters to send

        Returns
        -------
        out : result of the request
        """
        response = urllib.request.urlopen(
            self._api, urllib.parse.urlencode(params).encode('utf-8'))
        return response

    def _format_date(self, date):
        """Formats a date for Helioviewer API requests"""
        return parse_time(date).isot + "Z"

    def _get_source_id(self, key):
        """
        Returns source_id based on the key.
        """
        source_id_list = list(partial_key_match(key, self.data_sources))
        if len(source_id_list) != 1:
            raise KeyError(f"The values used: {key} do not correspond to one source_id "
                           f"but {len(source_id_list)} source_id(s)."
                           " Please check the list using HelioviewerClient.data_sources.")
        return source_id_list[0]



"""
============================================
Interacting with Data Using sunpy TimeSeries
============================================

This is an early run-through of the basic functionality of the sunpy TimeSeries
class.
This is intended primarily to demonstrate the current interface for discussion
of the final implementation. Much of the code will be changes as the class is
developed.
"""
import datetime
from collections import OrderedDict

import matplotlib.pyplot as plt
import numpy as np
from pandas import DataFrame

import astropy.units as u
from astropy.table import Table
from astropy.time import Time, TimeDelta

import sunpy.data.sample
import sunpy.timeseries
from sunpy.net import Fido
from sunpy.net import attrs as a
from sunpy.time import TimeRange, parse_time
from sunpy.util.metadata import MetaDict

##############################################################################
# Creating a TimeSeries from a file can be done using the factory.

ts_eve = sunpy.timeseries.TimeSeries(sunpy.data.sample.EVE_TIMESERIES, source='EVE')
ts_goes = sunpy.timeseries.TimeSeries(sunpy.data.sample.GOES_XRS_TIMESERIES, source='XRS')
ts_lyra = sunpy.timeseries.TimeSeries(sunpy.data.sample.LYRA_LEVEL3_TIMESERIES, source='LYRA')
ts_norh = sunpy.timeseries.TimeSeries(sunpy.data.sample.NORH_TIMESERIES, source='NoRH')
ts_rhessi = sunpy.timeseries.TimeSeries(sunpy.data.sample.RHESSI_TIMESERIES, source='RHESSI')
ts_gbm = sunpy.timeseries.TimeSeries(sunpy.data.sample.GBM_TIMESERIES, source='GBMSummary')

##############################################################################
# You can create a list of TimeSeries objects by using multiple files. First
# however, we shall download these files using `~sunpy.net.Fido`.

goes = Fido.search(a.Time("2012/06/01", "2012/06/04"), a.Instrument.xrs)
goes_files = Fido.fetch(goes)

# Using these new files you get a list:
lis_goes_ts = sunpy.timeseries.TimeSeries(goes_files[:2], source='XRS')
lis_goes_ts = sunpy.timeseries.TimeSeries(goes_files, source='XRS')
# Using concatenate=True kwarg you can merge the files into one TimeSeries:
combined_goes_ts = sunpy.timeseries.TimeSeries(goes_files, source='XRS', concatenate=True)
fig, ax = plt.subplots()

combined_goes_ts.plot()

##############################################################################
# You can concatenate manually:

combined_goes_ts = lis_goes_ts[0].concatenate(lis_goes_ts[1])
fig, ax = plt.subplots()
combined_goes_ts.plot()

##############################################################################
# The TimeSeries object has 3 primary data storage components:
#
# data: to get the underlying data, use to_dataframe()
# meta (OrderedDict): stores the metadata (like the Map)
# units (OrderedDict): stores the units for each column, with keys that match
# the name of each column.

ts_lyra.to_dataframe()
print(ts_lyra.meta)
print(ts_lyra.units)
# There are a couple of other useful properties you can quickly get:
print(ts_lyra.time_range)
print(ts_lyra.index)
print(ts_lyra.columns)

# Further data is available from within the metadata, you can filter out for a
# key using the TimeSeriesMetaData.get() method
combined_goes_ts.meta.get('telescop')
# This returns a TimeSeriesMetaData object, to get a list of just the
# values for this key use the values property of the metadata
combined_goes_ts.meta.get('telescop').values()

##############################################################################
# The ID used in the data Pandas DataFrame object will be a datetime, as can
# be seen using ts_lyra.index.
# You can access a specific value within the TimeSeries data DataFrame using
# all the normal Pandas methods.
# For example, the row with the index of 2015-01-01 00:02:00.008000:

ts_lyra.to_dataframe().loc[parse_time('2011-06-07 00:02:00.010').datetime]
# Pandas will actually parse a string to a datetime automatically if it can:
ts_lyra.to_dataframe().loc['2011-06-07 00:02:00.010']
# Pandas includes methods to find the indexes of the max/min values in a dataframe:
lyra_ch1_max_index = ts_lyra.to_dataframe()['CHANNEL1'].idxmax()
lyra_ch1_min_index = ts_lyra.to_dataframe()['CHANNEL1'].idxmin()

##############################################################################
# The TimeSeriesMetaData can be summarized:

combined_goes_ts.meta
print(combined_goes_ts.meta)
print(combined_goes_ts.meta.to_string(2))

##############################################################################
# The TimeSeries objects can be visualized using peek():

fig, ax = plt.subplots()
ts_goes.plot()
# And you can use subplots:
ts_eve.peek(subplots=True)

##############################################################################
# An individual column can be extracted from a TimeSeries:

ts_eve_extract = ts_eve.extract('CMLon')

##############################################################################
# You can truncate a TimeSeries using the truncate() method.
# This can use string datetime arguments, a sunpy TimeRange or integer value
# arguments (similar to slicing, but using function notation).
# Using integers we can get every other entry using:

ts_goes_trunc = ts_goes.truncate(0, 100000, 2)
# Or using a TimeRange:
tr = TimeRange('2011-06-07 05:00', '2011-06-07 06:30')
ts_goes_trunc = ts_goes.truncate(tr)
# Or using strings:
ts_goes_trunc = ts_goes.truncate('2011-06-07 05:00', '2011-06-07 06:30')
fig, ax = plt.subplots()
ts_goes_trunc.plot()

##############################################################################
# You can use Pandas resample method, for example to downsample.
# You can use 'mean', 'sum' and 'std' methods and any other methods in Pandas.
# Changing values within the datframe directly will often affect the units
# involved, but these won't be picked up by the TimeSeries object. Take care
# when doing this to ensure dimensional consistency.

df_downsampled = ts_goes_trunc.to_dataframe().resample('10T').mean()
ts_downsampled = sunpy.timeseries.TimeSeries(df_downsampled,
                                             ts_goes_trunc.meta,
                                             ts_goes_trunc.units)
fig, ax = plt.subplots()
ts_downsampled.plot()

##############################################################################
# Similarly, to upsample:

df_upsampled = ts_downsampled.to_dataframe().resample('1T').ffill()
# And this can be made into a TimeSeries using
ts_upsampled = sunpy.timeseries.TimeSeries(df_upsampled,
                                           ts_downsampled.meta,
                                           ts_downsampled.units)
fig, ax = plt.subplots()
ts_upsampled.plot()

plt.show()

##############################################################################
# The data from the TimeSeries can be retrieved in a number of formats:

ts_goes.to_dataframe()
ts_goes.to_table()
ts_goes.to_array()

##############################################################################
# Creating a TimeSeries from scratch can be done in a lot of ways, much like a
# Map.
# Input data can be in the form of a Pandas DataFrame (preferred), an astropy
# table or a numpy array.
# To generate some data and the corresponding dates

base = datetime.datetime.today()
dates = Time(base) - TimeDelta(np.arange(24 * 60)*u.minute)
intensity = np.sin(np.arange(0, 12 * np.pi, ((12 * np.pi) / (24 * 60))))
# Create the data DataFrame, header MetaDict and units OrderedDict
data = DataFrame(intensity, index=dates, columns=['intensity'])
units = OrderedDict([('intensity', u.W / u.m**2)])
meta = MetaDict({'key': 'value'})
# Create the time series
ts_custom = sunpy.timeseries.TimeSeries(data, meta, units)

# A more manual dataset would be a numpy array, which we can creat using:
tm = Time(['2000:002', '2001:345', '2002:345'])
a = [1, 4, 5]
b = [2.0, 5.0, 8.2]
c = ['x', 'y', 'z']
arr = np.stack([tm, a, b, c], axis=1)
# Note: this array needs to have the times in the first column, this can be in
# any form that can be converted using astropy.time.Time().

# We can use the array directly:
ts_from_arr = sunpy.timeseries.TimeSeries(arr, {})

# We can use this to create a table and even include units:
t = Table([tm, a, b, c], names=('time', 'a', 'b', 'c'), meta={'name': 'table'})
t['b'].unit = 's'  # Adding units
ts_from_table = sunpy.timeseries.TimeSeries(t, {})

# If you wanted to make a dataframe from this array then you could use:
df = DataFrame(data=arr[:, 1:])
df.index = tm
ts_from_df = sunpy.timeseries.TimeSeries(df, {})

##############################################################################
# You can optionally add units data, a dictionary matching column heading keys
# to an astropy unit.

units = OrderedDict([('a', u.Unit("ct")), ('b', u.Unit("ct")), ('c', u.Unit("ct"))])
ts_from_table = sunpy.timeseries.TimeSeries(t, {}, units)
ts_from_df = sunpy.timeseries.TimeSeries(df, {}, units)

##############################################################################
# Changing the units for a column simply requires changing the value:

ts_from_table.units['a'] = u.m

##############################################################################
# Quantities can be extracted from a column using the quantity(col_name) method:

colname = 'CMLat'
qua = ts_eve.quantity(colname)
print(qua)

##############################################################################
# You can add or overwrite a column using the add_column method.
# This method accepts an astropy quantity and will convert to the intended units
# if necessary.

qua_new = qua.value * 0.01 * ts_eve.units[colname]
print(qua_new)
ts_eve = ts_eve.add_column(colname, qua_new, overwrite=True)
# Otherwise you can also use a numpy array and it assume you're using the original
# units:
arr_new = (qua.value * 0.1 * ts_eve.units[colname]).value
ts_eve = ts_eve.add_column(colname, qua_new, overwrite=True)
# Finally, if you want to change the units used, you can specify a new unit for
# the column using the unit keyword:
qua_new = qua.value * 0.00001 * ts_eve.units[colname]
unit = u.W / (u.km**2)
ts_eve = ts_eve.add_column(colname, qua_new, unit=unit, overwrite=True)



"""
============================
Drawing a rectangle on a map
============================

This example will demonstrate how to draw a rectangle on a map using :meth:`~sunpy.map.GenericMap.draw_quadrangle`.
"""
import matplotlib.pyplot as plt

import astropy.units as u
from astropy.coordinates import SkyCoord

import sunpy.data.sample
import sunpy.map

################################################################################
# Let's start with a sample AIA image.

aia_map = sunpy.map.Map(sunpy.data.sample.AIA_171_IMAGE)

################################################################################
# Here are four different ways to draw a rectangle. The first three ways
# directly calls the `~astropy.coordinates.SkyCoord` class. The fourth way
# converts pixel coordinates to the equivalent `~astropy.coordinates.SkyCoord`
# objects using the :meth:`~sunpy.map.GenericMap.pixel_to_world`.

# sphinx_gallery_defer_figures

fig = plt.figure(figsize=(5, 5))
ax = fig.add_subplot(projection=aia_map)
aia_map.plot(axes=ax, clip_interval=(1, 99.99)*u.percent)

################################################################################
# Specify two opposite corners of the rectangle as a single, two-element
# SkyCoord object.

# sphinx_gallery_defer_figures

coords = SkyCoord(
    Tx=(100, 500) * u.arcsec,
    Ty=(200, 500) * u.arcsec,
    frame=aia_map.coordinate_frame,
)
aia_map.draw_quadrangle(
    coords,
    axes=ax,
    edgecolor="blue",
    linestyle="-",
    linewidth=2,
    label='2-element SkyCoord'
)

################################################################################
# Specify two opposite corners of the rectangle as separate SkyCoord objects.

# sphinx_gallery_defer_figures

bottom_left = SkyCoord(-500 * u.arcsec, 200 * u.arcsec, frame=aia_map.coordinate_frame)
top_right = SkyCoord(-100 * u.arcsec, 500 * u.arcsec, frame=aia_map.coordinate_frame)
aia_map.draw_quadrangle(
    bottom_left,
    axes=ax,
    top_right=top_right,
    edgecolor="green",
    linestyle="--",
    linewidth=2,
    label='two SkyCoords'
)

################################################################################
# Specify one corner of the rectangle and the rectangle's width and height.

# sphinx_gallery_defer_figures

bottom_left = SkyCoord(-500 * u.arcsec, -500 * u.arcsec, frame=aia_map.coordinate_frame)
width = 400 * u.arcsec
height = 300 * u.arcsec
aia_map.draw_quadrangle(
    bottom_left,
    axes=ax,
    width=width,
    height=height,
    edgecolor="yellow",
    linestyle="-.",
    linewidth=2,
    label='width/height'
)

################################################################################
# Draw a desired rectangle in pixel coordinates by first converting to SkyCoord objects.

bottom_left = aia_map.pixel_to_world(600 * u.pixel, 350 * u.pixel)
top_right = aia_map.pixel_to_world(800 * u.pixel, 450 * u.pixel)
aia_map.draw_quadrangle(
    bottom_left,
    axes=ax,
    top_right=top_right,
    edgecolor="red",
    linestyle=":",
    linewidth=2,
    label='pixel_to_world()'
)
ax.legend()

plt.show()



"""
=============================
Plotting a solar cycle index
=============================

This example demonstrates how to plot the solar cycle in terms of
the number of sunspots and a prediction for the next few years.
"""
import matplotlib.pyplot as plt

import astropy.units as u
from astropy.time import Time, TimeDelta

import sunpy.timeseries as ts
from sunpy.net import Fido
from sunpy.net import attrs as a
from sunpy.time import TimeRange

###############################################################################
# The U.S. Dept. of Commerce, NOAA, Space Weather Prediction Center (SWPC)
# provides recent solar cycle indices which includes different sunspot numbers,
# radio flux, and geomagnetic index. They also provide predictions for how the
# sunspot number and radio flux will evolve. Predicted values are based on the
# consensus of the Solar Cycle 24 Prediction Panel.
#
# We will first search for and then download the data.

time_range = TimeRange("2008-06-01 00:00", Time.now())
result = Fido.search(a.Time(time_range), a.Instrument('noaa-indices'))
f_noaa_indices = Fido.fetch(result)
result = Fido.search(a.Time(time_range.end, time_range.end + TimeDelta(4 * u.year)),
                     a.Instrument('noaa-predict'))
f_noaa_predict = Fido.fetch(result)

###############################################################################
#  We then load them into individual `~sunpy.timeseries.TimeSeries` objects.

noaa = ts.TimeSeries(f_noaa_indices, source='noaaindices').truncate(time_range)
noaa_predict = ts.TimeSeries(f_noaa_predict, source='noaapredictindices')

###############################################################################
# Finally, we plot both ``noaa`` and ``noaa_predict`` for the sunspot number.
# In this case we use the S.I.D.C. Brussels International Sunspot Number (RI).
# The predictions provide both a high and low values, which we plot below as
# ranges.

plt.figure()
plt.plot(noaa.index, noaa.quantity('sunspot RI'), label='Sunspot Number')
plt.plot(noaa_predict.index, noaa_predict.quantity('sunspot'),
         color='grey', label='Near-term Prediction')
plt.fill_between(noaa_predict.index, noaa_predict.quantity('sunspot low'),
                 noaa_predict.quantity('sunspot high'), alpha=0.3, color='grey')
plt.ylim(bottom=0)
plt.ylabel('Sunspot Number')
plt.xlabel('Year')
plt.legend()

plt.show()



import copy
from collections import OrderedDict

import pytest

from sunpy.time import TimeRange
from sunpy.timeseries import TimeSeriesMetaData
from sunpy.util import SunpyUserWarning
from sunpy.util.metadata import MetaDict

# =============================================================================
# Creating TimeSeriesMetaData Objects
# =============================================================================


@pytest.fixture
def basic_1_md():
    tr = TimeRange('2010-01-01 13:59:57.468999', '2010-01-02 13:59:56.091999')
    colnames = ['column1', 'column2']
    metadict = MetaDict(OrderedDict(
        [('md1_key1', 'value1'), ('md1_key2', 'value2'), ('all_same', 'value3'), ('all_different', 'diff_1')]))
    lis = [(tr, colnames, metadict)]
    return TimeSeriesMetaData(lis)


@pytest.fixture
def basic_2_md():
    tr = TimeRange('2010-01-02 13:59:57.468999', '2010-01-03 13:59:56.091999')
    colnames = ['column1', 'column2']
    metadict = MetaDict(OrderedDict(
        [('md2_key1', 'value1'), ('md2_key2', 'value2'), ('all_same', 'value3'), ('all_different', 'diff_2')]))
    lis = [(tr, colnames, metadict)]
    return TimeSeriesMetaData(lis)


@pytest.fixture
def basic_3_md():
    tr = TimeRange('2010-01-03 13:59:57.468999', '2010-01-03 13:59:56.091999')
    colnames = ['column1', 'column2']
    metadict = MetaDict(OrderedDict(
        [('md3_key1', 'value1'), ('md3_key2', 'value2'), ('all_same', 'value3'), ('all_different', 'diff_3')]))
    lis = [(tr, colnames, metadict)]
    return TimeSeriesMetaData(lis)


@pytest.fixture
def basic_4_md():
    tr = TimeRange('2010-01-01 20:59:57.468999', '2010-01-03 20:59:56.091999')
    colnames = ['md4_column1', 'md4_column2']
    metadict = MetaDict(OrderedDict(
        [('md4_key1', 'value1'), ('md4_key2', 'value2'), ('all_same', 'value3'), ('all_different', 'diff_4')]))
    tup = (tr, colnames, metadict)
    return TimeSeriesMetaData(tup)


@pytest.fixture
def overlap_and_interleave_with_basic_1_md():
    tr = TimeRange('2010-01-01 01:01:00.0', '2010-01-02 01:01:00.0')
    colnames = ['column1', 'column2']
    metadict = MetaDict(OrderedDict(
        [('other_key1', 'value1'), ('other_key2', 'value2'), ('all_same', 'value3'), ('all_different', 'diff_5')]))
    lis = [(tr, colnames, metadict)]
    return TimeSeriesMetaData(lis)

# =============================================================================
# Test Creating TimeSeriesMetaData With Limited Input
# =============================================================================


def test_create_mithout_metadata():
    tr = TimeRange('2010-01-01 13:59:57.468999', '2010-01-02 13:59:56.091999')
    colnames = ['column1', 'column2']
    tsmd_1 = TimeSeriesMetaData(timerange=tr, colnames=colnames)
    assert isinstance(tsmd_1, TimeSeriesMetaData)
    assert tsmd_1.metadata[0][1] == colnames
    with pytest.warns(SunpyUserWarning, match='No time range given for metadata'):
        tsmd_2 = TimeSeriesMetaData(timerange=tr)
    assert isinstance(tsmd_1, TimeSeriesMetaData)
    assert tsmd_2.metadata[0][1] == []
    assert tsmd_1.metadata[0][0] == tsmd_2.metadata[0][0] == tr
    assert tsmd_1.metadata[0][2] == tsmd_2.metadata[0][2] == MetaDict()
    assert len(tsmd_1.metadata) == len(tsmd_2.metadata) == 1


def test_create_mithout_metadata_or_timerange():
    # without a timerange we should get errors
    colnames = ['column1', 'column2']
    with pytest.raises(ValueError):
        TimeSeriesMetaData(colnames=colnames)
    with pytest.raises(ValueError):
        TimeSeriesMetaData()


# =============================================================================
# Test Appending TimeSeriesMetaData Objects
# =============================================================================

def test_append_similar(basic_1_md):
    appended = copy.deepcopy(basic_1_md)
    appended.append(*basic_1_md.metadata[0])
    # The duplicate should not have been added
    assert appended == basic_1_md


@pytest.fixture
def basic_ascending_append_md(basic_1_md, basic_2_md, basic_3_md):
    appended = copy.deepcopy(basic_1_md)
    appended.append(*basic_2_md.metadata[0])
    appended.append(*basic_3_md.metadata[0])
    return appended


def test_basic_ascending_append_md(basic_1_md, basic_2_md, basic_3_md, basic_ascending_append_md):
    # Check all the entries are in the correct order
    assert basic_ascending_append_md.metadata[0] == basic_1_md.metadata[0]
    assert basic_ascending_append_md.metadata[1] == basic_2_md.metadata[0]
    assert basic_ascending_append_md.metadata[2] == basic_3_md.metadata[0]


def test_basic_descending_append_md(basic_1_md, basic_2_md, basic_3_md, basic_ascending_append_md):
    appended = copy.deepcopy(basic_3_md)
    appended.append(*basic_1_md.metadata[0])
    appended.append(*basic_2_md.metadata[0])
    assert appended == basic_ascending_append_md


def test_basic_random_append_md(basic_1_md, basic_2_md, basic_3_md, basic_ascending_append_md):
    appended = copy.deepcopy(basic_3_md)
    appended.append(*basic_1_md.metadata[0])
    appended.append(*basic_2_md.metadata[0])
    assert appended == basic_ascending_append_md


@pytest.fixture
def complex_append_md(basic_1_md, basic_2_md, basic_3_md, basic_4_md):
    appended = copy.deepcopy(basic_1_md)
    appended.append(*basic_2_md.metadata[0])
    appended.append(*basic_3_md.metadata[0])
    appended.append(*basic_4_md.metadata[0])
    return appended


def test_complex_append_md(basic_1_md, basic_2_md, basic_3_md, basic_4_md, complex_append_md):
    # Check all the entries are in the correct order
    assert complex_append_md.metadata[0] == basic_1_md.metadata[0]
    assert complex_append_md.metadata[1] == basic_4_md.metadata[0]
    assert complex_append_md.metadata[2] == basic_2_md.metadata[0]
    assert complex_append_md.metadata[3] == basic_3_md.metadata[0]


def test_append_invalid_timerange(basic_1_md):
    appended = copy.deepcopy(basic_1_md)
    with pytest.raises(ValueError):
        appended.append('not_a_timerange', basic_1_md.metadata[0][1], basic_1_md.metadata[0][2])


# =============================================================================
# Test TimeSeriesMetaData Concatenation
# =============================================================================

def test_concatenate_ts(basic_ascending_append_md, basic_4_md, complex_append_md):
    concatenated = copy.deepcopy(basic_ascending_append_md)
    concatenated = concatenated.concatenate(basic_4_md)
    assert concatenated == complex_append_md


def test_concatenate_list(basic_ascending_append_md, basic_3_md, basic_4_md, complex_append_md):
    concatenated = copy.deepcopy(basic_ascending_append_md)
    concatenated = concatenated.concatenate([basic_3_md, basic_4_md])
    assert concatenated == complex_append_md


def test_concatenate_invalid_type(basic_ascending_append_md):
    concatenated = copy.deepcopy(basic_ascending_append_md)
    with pytest.raises(TypeError, match="Invalid type provided: <class 'int'>. "
                                        "Please provide a TimeSeriesMetaData object or "
                                        "an iterable containing TimeSeriesMetaData objects."):
        concatenated = concatenated.concatenate(100)


def test_concatenate_invalid_type_within_list(basic_ascending_append_md):
    concatenated = copy.deepcopy(basic_ascending_append_md)
    with pytest.raises(TypeError, match="Invalid type within iterable. Iterable must only contain "
                                        "TimeSeriesMetaData objects."):
        concatenated = concatenated.concatenate([100])


# =============================================================================
# Test TimeSeriesMetaData Truncation
# =============================================================================

@pytest.fixture
def truncated_none_md(basic_ascending_append_md):
    # This timerange covers the whole range of metadata, so no change is expected
    tr = TimeRange('2010-01-01 1:59:57.468999', '2010-01-04 23:59:56.091999')
    truncated = copy.deepcopy(basic_ascending_append_md)
    truncated._truncate(tr)
    return truncated


@pytest.fixture
def truncated_start_md(basic_ascending_append_md):
    # This time range starts after the original, so expect truncation
    tr = TimeRange('2010-01-02 20:59:57.468999', '2010-01-04 23:59:56.091999')
    truncated = copy.deepcopy(basic_ascending_append_md)
    truncated._truncate(tr)
    return truncated


@pytest.fixture
def truncated_end_md(basic_ascending_append_md):
    # This time range ends before the original, so expect truncation
    tr = TimeRange('2010-01-01 1:59:57.468999', '2010-01-03 1:59:56.091999')
    truncated = copy.deepcopy(basic_ascending_append_md)
    truncated._truncate(tr)
    return truncated


@pytest.fixture
def truncated_both_md(basic_ascending_append_md):
    # This time range starts after and ends before the original, so expect truncation
    tr = TimeRange('2010-01-02 20:59:57.468999', '2010-01-03 1:59:56.091999')
    truncated = copy.deepcopy(basic_ascending_append_md)
    truncated._truncate(tr)
    return truncated


@pytest.fixture
def truncated_new_tr_all_before_md(basic_ascending_append_md):
    # Time range begins and ends before the data
    tr = TimeRange('2010-01-01 01:01:01.000000', '2010-01-01 02:01:01.000000')
    truncated = copy.deepcopy(basic_ascending_append_md)
    truncated._truncate(tr)
    return truncated


@pytest.fixture
def truncated_new_tr_all_after_md(basic_ascending_append_md):
    # Time range begins and ends after the data
    tr = TimeRange('2010-01-04 01:01:01.000000', '2010-01-04 02:01:01.000000')
    truncated = copy.deepcopy(basic_ascending_append_md)
    truncated._truncate(tr)
    return truncated


# =============================================================================
# Test TimeSeriesMetaData TimeRanges
# =============================================================================

def test_truncated_none_tr(basic_ascending_append_md, truncated_none_md):
    assert basic_ascending_append_md.time_range == truncated_none_md.time_range


def test_truncated_start_tr(truncated_start_md):
    tr = TimeRange('2010-01-02 20:59:57.468999', truncated_start_md.time_range.end)
    assert truncated_start_md.time_range == tr


def test_truncated_end_tr(basic_ascending_append_md, truncated_end_md):
    tr = TimeRange(truncated_end_md.time_range.start, '2010-01-03 1:59:56.091999')
    assert truncated_end_md.time_range == tr


def test_truncated_both_tr(truncated_both_md):
    tr = TimeRange('2010-01-02 20:59:57.468999', '2010-01-03 1:59:56.091999')
    assert truncated_both_md.time_range == tr


def test_truncated_tr_outside(truncated_new_tr_all_before_md, truncated_new_tr_all_after_md):
    assert truncated_new_tr_all_before_md.metadata == truncated_new_tr_all_after_md.metadata == []


def test_basic_ascending_append_tr(basic_1_md, basic_3_md, basic_ascending_append_md):
    tr = TimeRange(basic_1_md.time_range.start, basic_3_md.time_range.end)
    assert basic_ascending_append_md.time_range == tr


def test_complex_append_tr(basic_1_md, basic_4_md, complex_append_md):
    tr = TimeRange(basic_1_md.time_range.start, basic_4_md.time_range.end)
    assert complex_append_md.time_range == tr


# =============================================================================
# Test TimeSeriesMetaData find method
# =============================================================================

def test_find_return_type(basic_ascending_append_md):
    assert isinstance(basic_ascending_append_md.find(), TimeSeriesMetaData)


def test_find_no_filters(basic_ascending_append_md, basic_1_md):
    assert basic_ascending_append_md.find() == basic_ascending_append_md


def test_find_time_filter(basic_1_md, basic_2_md, basic_4_md, complex_append_md):
    assert complex_append_md.find(time='2010-01-01 14:59:57.468999') == basic_1_md
    temp_md = copy.deepcopy(basic_2_md)
    temp_md = temp_md.concatenate(basic_4_md)
    assert complex_append_md.find(time='2010-01-02 20:59:57.468999') == temp_md


def test_find_colname_filter(basic_4_md, complex_append_md, basic_ascending_append_md):
    assert complex_append_md.find(colname='md4_column2') == basic_4_md
    assert complex_append_md.find(colname='column1') == basic_ascending_append_md


def test_find_both_filters(basic_2_md, basic_4_md, complex_append_md):
    assert complex_append_md.find(time='2010-01-02 20:59:57.468999',
                                  colname='column2') == basic_2_md
    assert complex_append_md.find(time='2010-01-02 20:59:57.468999',
                                  colname='md4_column1') == basic_4_md


# =============================================================================
# Test TimeSeriesMetaData get and update methods
# =============================================================================

def test_get_return_type(complex_append_md):
    assert isinstance(complex_append_md.get('md1_key1'), TimeSeriesMetaData)


def test_get_no_filters(complex_append_md):
    assert complex_append_md.get('md1_key1').values() == ['value1']
    assert complex_append_md.get('all_same').values() == ['value3']


def test_get_time_filter(complex_append_md):
    assert complex_append_md.get(
        'md1_key1', time='2010-01-01 20:59:57.468999').values() == ['value1']
    assert complex_append_md.get(
        'md2_key2', time='2010-01-02 20:59:57.468999').values() == ['value2']
    assert complex_append_md.get(
        'all_same', time='2010-01-01 20:59:57.468999').values() == ['value3']
    assert complex_append_md.get(
        'all_different', time='2010-01-01 20:59:57.468999').values() == ['diff_1', 'diff_4']


def test_get_colname_filter(complex_append_md):
    assert complex_append_md.get('md1_key1', colname='column1').values() == ['value1']
    assert complex_append_md.get('md2_key2', colname='column2').values() == ['value2']
    assert complex_append_md.get('all_same', colname='column1').values() == ['value3']
    assert complex_append_md.get('all_different', colname='column1').values() == [
        'diff_1', 'diff_2', 'diff_3']


def test_get_both_filters(complex_append_md):
    assert complex_append_md.get(
        'all_different', time='2010-01-02 20:59:57.468999', colname='column2').values() == ['diff_2']


# =============================================================================
# Test TimeSeriesMetaData update method
# =============================================================================

def test_update(complex_append_md):
    assert isinstance(complex_append_md.get('md1_key1'), TimeSeriesMetaData)


def test_update_all_dict_type_input(complex_append_md):
    # Check all three dictionary types work the same
    updated_dict_md = copy.deepcopy(complex_append_md)
    updated_dict_md.update({'added': 'added'})
    updated_ordereddict_md = copy.deepcopy(complex_append_md)
    updated_ordereddict_md.update(MetaDict(OrderedDict([('added', 'added')])))
    updated_metadict_md = copy.deepcopy(complex_append_md)
    updated_metadict_md.update(MetaDict(OrderedDict([('added', 'added')])))
    assert updated_dict_md == updated_ordereddict_md == updated_metadict_md


def test_update_overwrite(complex_append_md, overwrite=True):
    updated_not_overwritten_md = copy.deepcopy(complex_append_md)
    updated_not_overwritten_md.update({'all_same': 'updated'})
    updated_overwritten_md = copy.deepcopy(complex_append_md)
    updated_overwritten_md.update({'all_same': 'updated'}, overwrite=True)
    assert updated_not_overwritten_md == complex_append_md
    assert updated_overwritten_md.get('all_same').values() == ['updated']


def test_update_time_filter(complex_append_md):
    updated_md = copy.deepcopy(complex_append_md)
    updated_md.update({'all_same': 'updated'}, time='2010-01-01 20:59:57.468999', overwrite=True)
    assert updated_md.metadata[0][2]['all_same'] == updated_md.metadata[1][2]['all_same'] == 'updated'
    assert updated_md.metadata[2][2]['all_same'] == updated_md.metadata[3][2]['all_same'] == 'value3'


def test_update_colname_filter(complex_append_md):
    updated_md = copy.deepcopy(complex_append_md)
    updated_md.update({'all_same': 'updated'}, colname='column1', overwrite=True)
    assert updated_md.metadata[0][2]['all_same'] == updated_md.metadata[2][2]['all_same'] == updated_md.metadata[3][2]['all_same'] == 'updated'
    assert updated_md.metadata[1][2]['all_same'] == 'value3'


def test_update_both_filters(complex_append_md):
    updated_md = copy.deepcopy(complex_append_md)
    updated_md.update({'all_same': 'updated'}, colname='column1',
                      time='2010-01-01 23:59:57.468999', overwrite=True)
    assert updated_md.metadata[0][2]['all_same'] == 'updated'
    assert updated_md.metadata[1][2]['all_same'] == updated_md.metadata[2][2]['all_same'] == updated_md.metadata[3][2]['all_same'] == 'value3'


# =============================================================================
# Test Misc Methods
# =============================================================================

def test_get_index(basic_1_md, basic_ascending_append_md):
    assert basic_ascending_append_md.get_index(0) == basic_1_md.get_index(0)


def test_equality(basic_1_md, basic_2_md, basic_ascending_append_md):
    basic_1_copy_md = copy.deepcopy(basic_1_md)
    assert basic_1_md == basic_1_copy_md
    assert basic_1_md != basic_2_md
    assert basic_1_md != basic_ascending_append_md


def test_to_string_basic(basic_1_md):
    default_str = basic_1_md.to_string()
    assert isinstance(default_str, str)

    # check this matches the __str__ and __repr__ methods
    default_str == basic_1_md.__str__() == basic_1_md.__repr__()


def test_to_string_depth(basic_1_md):
    depth_1_str = basic_1_md.to_string(depth=1)
    assert len(depth_1_str.split('\n')) == 7
    assert len(depth_1_str.split('...')) == 4
    assert basic_1_md.metadata[0][1][0] in depth_1_str

    # for depth of 2 the time-range will take exactly 2 lines
    depth_2_str = basic_1_md.to_string(depth=2)
    assert len(depth_2_str.split('\n')) == 8
    assert len(depth_2_str.split('...')) == 2
    assert (basic_1_md.metadata[0][1][0] in depth_2_str) and (
        basic_1_md.metadata[0][1][1] in depth_2_str)


def test_to_string_width(basic_1_md):
    width_110_str = basic_1_md.to_string(width=110)
    split = width_110_str.split('\n')
    assert len(split[0]) == len(split[1]) == len(split[2]) == len(split[3]) == 110

    width_60_str = basic_1_md.to_string(width=60)
    split = width_60_str.split('\n')
    assert len(split[0]) == len(split[1]) == len(split[2]) == len(split[3]) == 60


def test_to_string_few_metadict_entries(basic_1_md):
    tr = basic_1_md.metadata[0][0]
    colnames = basic_1_md.metadata[0][1]
    metadict = MetaDict(OrderedDict([('md1_key1', 'value1')]))
    lis = [(tr, colnames, metadict)]
    basic_1_md_less_metadict_entries = TimeSeriesMetaData(lis)
    depth_3_str = basic_1_md_less_metadict_entries.to_string(depth=3)
    assert isinstance(depth_3_str, str)


def test_timeranges(basic_ascending_append_md):
    lis = []
    lis.append(basic_ascending_append_md.metadata[0][0])
    lis.append(basic_ascending_append_md.metadata[1][0])
    lis.append(basic_ascending_append_md.metadata[2][0])
    assert basic_ascending_append_md.timeranges == lis


def test_columns(complex_append_md):
    lis = complex_append_md.metadata[0][1] + complex_append_md.metadata[1][1]
    lis.sort()
    assert complex_append_md.columns == lis


def test_metas(complex_append_md):
    lis = []
    lis.append(complex_append_md.metadata[0][2])
    lis.append(complex_append_md.metadata[1][2])
    lis.append(complex_append_md.metadata[2][2])
    lis.append(complex_append_md.metadata[3][2])
    assert complex_append_md.metas == lis


def test_rename_column(complex_append_md):
    col_renamed_md = copy.deepcopy(complex_append_md)
    old = complex_append_md.metadata[0][1][0]
    new = 'renamed'
    col_renamed_md._rename_column(old, new)
    assert col_renamed_md.metadata[0][1][0] == col_renamed_md.metadata[2][1][0] == col_renamed_md.metadata[3][1][0] == new


def test_remove_column(complex_append_md):
    col_removed_md = copy.deepcopy(complex_append_md)
    col = complex_append_md.metadata[0][1][0]
    col_removed_md._remove_columns(col)
    assert col_removed_md.metadata[0][1] == col_removed_md.metadata[2][1] == col_removed_md.metadata[3][1] == [
        complex_append_md.metadata[0][1][1]]
    assert col_removed_md.metadata[1][1] == [
        complex_append_md.metadata[1][1][0], complex_append_md.metadata[1][1][1]]


def test_remove_columns(complex_append_md):
    cols_removed_md = copy.deepcopy(complex_append_md)
    lis = [complex_append_md.metadata[0][1][0], complex_append_md.metadata[1][1][1]]
    cols_removed_md._remove_columns(lis)
    assert cols_removed_md.metadata[0][1] == cols_removed_md.metadata[2][1] == cols_removed_md.metadata[3][1] == [
        complex_append_md.metadata[0][1][1]]
    assert cols_removed_md.metadata[1][1] == [complex_append_md.metadata[1][1][0]]


def test_validate_meta_good(complex_append_md):
    assert complex_append_md._validate_meta(complex_append_md)


def test_validate_meta_interleaved(basic_1_md, overlap_and_interleave_with_basic_1_md):
    concatenated = copy.deepcopy(basic_1_md)
    concatenated = concatenated.concatenate(overlap_and_interleave_with_basic_1_md)
    with pytest.warns(SunpyUserWarning, match='Metadata entries 0 and 1 contain interleaved data.'):
        assert concatenated._validate_meta(concatenated)



from sunpy.net.attr import SimpleAttr

__all__ = ["SatelliteNumber"]


# Define a custom __dir__ to restrict tab-completion to __all__
def __dir__():
    return __all__


class SatelliteNumber(SimpleAttr):
    """
    The GOES Satellite Number
    """



"""
This module provides a web scraper.
"""
from sunpy.net.scraper import Scraper
from sunpy.util.decorators import deprecated

__all__ = ['Scraper']


@deprecated(since="3.1", alternative="sunpy.net.Scraper")
class Scraper(Scraper):
    pass



"""
This module provides a FITS file reader.

.. warning::

   ``sunpy.io.fits`` is deprecated, and will be removed in sunpy 4.1. This is
   because it was designed for interal use only. We instead recommend users
   use the `astropy.io.fits` module, which provides more generic functionality
   to read FITS files.


Notes
-----

1. FITS files allow comments to be attached to every value in the header.
   This is implemented in this module as a KEYCOMMENTS dictionary in the
   sunpy header. To add a comment to the file on write, add a comment to this
   dictionary with the same name as a key in the header (upcased).

2. Due to the way `~astropy.io.fits` works with images, the header dictionary may
   differ depending on whether is accessed before or after the fits[0].data
   is requested. If the header is read before the data then the original
   header will be returned. If the header is read after the data has been
   accessed then the data will have been scaled and a modified header
   reflecting these changes will be returned: BITPIX may differ and
   BSCALE and B_ZERO may be dropped in the modified version.

3. The verify('silentfix+warn') call attempts to handle violations of the FITS
   standard. For example, ``nan`` values will be converted to "nan" strings.
   Attempting to cast a `astropy.io.fits.Header` to a dictionary while it contains
   invalid header tags will result in an error so verifying it early on
   makes the header easier to work with later.
"""
import os
import re
import sys
import math
import traceback
import collections
import collections.abc

from astropy.io import fits

from sunpy.io.header import FileHeader
from sunpy.util.exceptions import warn_metadata, warn_user

__all__ = ['header_to_fits', 'read', 'get_header', 'write', 'extract_waveunit']

HDPair = collections.namedtuple('HDPair', ['data', 'header'])


def read(filepath, hdus=None, memmap=None, **kwargs):
    """
    Read a fits file.

    Parameters
    ----------
    filepath : `str`
        The fits file to be read.
    hdus : `int` or iterable
        The HDU indexes to read from the file.
    **kwargs : `dict`, optional
        Passed to `astropy.io.fits.open`.

    Returns
    -------
    pairs : `list`
        A list of (data, header) tuples

    Notes
    -----
    This routine reads all the HDU's in a fits file and returns a list of the
    data and a FileHeader instance for each one.

    Also all comments in the original file are concatenated into a single
    "comment" key in the returned FileHeader.
    """
    with fits.open(filepath, ignore_blank=True, memmap=memmap, **kwargs) as hdulist:
        if hdus is not None:
            if isinstance(hdus, int):
                hdulist = hdulist[hdus]
            elif isinstance(hdus, collections.abc.Iterable):
                hdulist = [hdulist[i] for i in hdus]

        hdulist = fits.hdu.HDUList(hdulist)
        for h in hdulist:
            h.verify('silentfix+warn')

        headers = get_header(hdulist)
        pairs = []

        for i, (hdu, header) in enumerate(zip(hdulist, headers)):
            try:
                pairs.append(HDPair(hdu.data, header))
            except (KeyError, ValueError) as e:
                message = f"Error when reading HDU {i}. Skipping.\n"
                for line in traceback.format_tb(sys.exc_info()[2]):
                    message += line
                    message += '\n'
                message += repr(e)
                warn_user(message)

    return pairs


def get_header(afile):
    """
    Read a fits file and return just the headers for all HDU's.

    Parameters
    ----------
    afile : `str` or `astropy.io.fits.HDUList`
        The file to be read, or HDUList to process.

    Returns
    -------
    headers : `list`
        A list of `sunpy.io.header.FileHeader` headers.
    """
    if isinstance(afile, fits.HDUList):
        hdulist = afile
        close = False
    else:
        hdulist = fits.open(afile, ignore_blank=True)
        hdulist.verify('silentfix')
        close = True

    try:
        headers = []
        for hdu in hdulist:
            try:
                comment = "".join(hdu.header['COMMENT']).strip()
            except KeyError:
                comment = ""
            try:
                history = "".join(hdu.header['HISTORY']).strip()
            except KeyError:
                history = ""

            header = FileHeader(hdu.header)
            header['COMMENT'] = comment
            header['HISTORY'] = history

            # Strip out KEYCOMMENTS to a dict, the hard way
            keydict = {}
            for card in hdu.header.cards:
                if card.comment != '':
                    keydict.update({card.keyword: card.comment})
            header['KEYCOMMENTS'] = keydict

            headers.append(header)
    finally:
        if close:
            hdulist.close()
    return headers


def write(fname, data, header, hdu_type=None, **kwargs):
    """
    Take a data header pair and write a FITS file.

    Parameters
    ----------
    fname : `str`
        File name, with extension.
    data : `numpy.ndarray`
        n-dimensional data array.
    header : `dict`
        A header dictionary.
    hdu_type : `~astropy.io.fits.hdu.base.ExtensionHDU` instance or class, optional
        By default, a FITS file is written with the map in its primary HDU.
        If a type is given, a new HDU of this type will be created.
        If a HDU instance is given, its data and header will be updated from the map.
        Then that HDU instance will be written to the file.
    kwargs :
        Additional keyword arguments are given to
        `~astropy.io.fits.HDUList.writeto`.
    """
    # Copy header so the one in memory is left alone while changing it for
    # write.
    header = header.copy()

    fits_header = header_to_fits(header)

    if isinstance(fname, str):
        fname = os.path.expanduser(fname)

    fitskwargs = {'output_verify': 'fix'}
    fitskwargs.update(kwargs)

    if not hdu_type:
        hdu_type = fits.PrimaryHDU

    if isinstance(hdu_type, (fits.PrimaryHDU, fits.hdu.base.ExtensionHDU)):
        hdu = hdu_type  # HDU already initialised

        # Merge `header` into HDU's header
        # Values in `header` take priority, including cards such as
        # 'SIMPLE' and 'BITPIX'.
        hdu.header.extend(fits_header, strip=False, update=True)

        # Set the HDU's data
        hdu.data = data
    else:
        hdu = hdu_type(data=data, header=fits_header)

    if not isinstance(hdu, fits.PrimaryHDU):
        hdul = fits.HDUList([fits.PrimaryHDU(), hdu])
    else:
        hdul = fits.HDUList([hdu])

    hdul.writeto(fname, **fitskwargs)


def header_to_fits(header):
    """
    Convert a header dict to a `~astropy.io.fits.Header`.
    """
    # Copy the header to avoid modifying it in place
    header = header.copy()
    # The comments need to be added to the header separately from the normal
    # kwargs. Find and deal with them:
    fits_header = fits.Header()
    # Check Header
    key_comments = header.pop('KEYCOMMENTS', False)

    for k, v in header.items():
        # Drop any keys that have non-ascii characters
        if not fits.Card._ascii_text_re.match(str(v)):
            warn_metadata(f'The meta key {k} is not valid ascii, dropping from the FITS header')
            continue
        # Drop any keys which are too long to save into FITS
        if len(k) > 8:
            warn_metadata(f"The meta key {k} is too long, dropping from the FITS header "
                          "(maximum allowed key length is 8 characters).")
            continue

        if isinstance(v, float) and math.isnan(v):
            warn_metadata(f'The meta key {k} has a NaN value, which is not valid in a FITS '
                          'header, dropping from the FITS header')
            continue

        if k.upper() in ('COMMENT', 'HV_COMMENT'):
            comments = str(v).split('\n')
            for com in comments:
                fits_header.add_comment(com)
        elif k.upper() == 'HISTORY':
            hists = str(v).split('\n')
            for hist in hists:
                fits_header.add_history(hist)
        elif isinstance(v, fits.header._HeaderCommentaryCards):
            if k != '':
                fits_header.append(fits.Card(k, str(v).split('\n')))
        else:
            # For some horrific reason, we save a list to the wavelnth key in
            # sources/rhessi.py. This is the least invasive fix for that stupidity.
            if isinstance(v, list):
                v = str(v)
            fits_header.append(fits.Card(k, v))

    if isinstance(key_comments, dict):
        for k, v in key_comments.items():
            # Check that the Card for the comment exists before trying to write to it.
            if k in fits_header:
                fits_header.comments[k] = v
    elif key_comments:

        raise TypeError("KEYCOMMENTS must be a dictionary")

    return fits_header


def extract_waveunit(header):
    """
    Attempt to read the wavelength unit from a given FITS header.

    Parameters
    ----------
    header : `sunpy.io.header.FileHeader`
        One `~sunpy.io.header.FileHeader` instance which was created by
        reading a FITS file. For example, `sunpy.io.fits.get_header` returns a list of
        such instances.

    Returns
    -------
    waveunit : `str`
        The wavelength unit that could be found or ``None`` otherwise.

    Examples
    --------
    The goal of this function is to return a string that can be used in
    conjunction with the astropy.units module so that the return value can be
    directly passed to `astropy.units.Unit`.

    >>> import astropy.units
    >>> header = {'WAVEUNIT': 'Angstrom', 'KEYCOMMENTS': {}}
    >>> waveunit = extract_waveunit(header)
    >>> if waveunit is not None:
    ...     unit = astropy.units.Unit(waveunit)
    """
    # algorithm: try the following procedures in the following order and return
    # as soon as a waveunit could be detected
    # 1. read header('WAVEUNIT'). If None, go to step 2.
    # 1.1 -9 -> 'nm'
    # 1.2 -10 -> 'angstrom'
    # 1.3 0 -> go to step 2
    # 1.4 if neither of the above, return the value itself in lowercase
    # 2. parse waveunit_comment
    # 2.1 'in meters' -> 'm'
    # 3. parse wavelnth_comment
    # 3.1 "[$UNIT] ..." -> $UNIT
    # 3.2 "Observed wavelength ($UNIT)" -> $UNIT
    def parse_waveunit_comment(waveunit_comment):
        if waveunit_comment == 'in meters':
            return 'm'

    if 'KEYCOMMENTS' not in header:
        return None
    waveunit_comment = header['KEYCOMMENTS'].get('WAVEUNIT')
    wavelnth_comment = header['KEYCOMMENTS'].get('WAVELNTH')
    waveunit = header.get('WAVEUNIT')
    if waveunit is not None:
        metre_submultiples = {
            0: parse_waveunit_comment(waveunit_comment),
            -1: 'dm',
            -2: 'cm',
            -3: 'mm',
            -6: 'um',
            -9: 'nm',
            -10: 'angstrom',
            -12: 'pm',
            -15: 'fm',
            -18: 'am',
            -21: 'zm',
            -24: 'ym'}
        waveunit = metre_submultiples.get(waveunit, str(waveunit).lower())
    elif waveunit_comment is not None:
        waveunit = parse_waveunit_comment(waveunit_comment)
    elif wavelnth_comment is not None:
        # supported formats (where $UNIT is the unit like "nm" or "Angstrom"):
        #   "Observed wavelength ($UNIT)"
        #   "[$UNIT] ..."
        parentheses_pattern = r'Observed wavelength \((\w+?)\)$'
        brackets_pattern = r'^\[(\w+?)\]'
        for pattern in [parentheses_pattern, brackets_pattern]:
            m = re.search(pattern, wavelnth_comment)
            if m is not None:
                waveunit = m.group(1)
                break
    if waveunit == '':
        return None  # To fix problems associated with HMI FITS.
    return waveunit



"""Test cases for SOHO Map subclasses.
This particular test file pertains to LASCOMap.
@Author: Pritish C. (VaticanCameos)
"""

import os
import glob
from textwrap import dedent

import numpy as np
import pytest

import astropy.units as u
from astropy.io import fits

import sunpy.data.test
from sunpy.map import Map
from sunpy.map.sources.soho import LASCOMap
from sunpy.tests.helpers import skip_glymur
from sunpy.time import parse_time
from sunpy.util.exceptions import SunpyMetadataWarning

path = sunpy.data.test.rootdir


def lascoc3():
    # Please do not edit this header, it is a 1:1 copy of a LASCO C3 header
    raw_header = dedent("""\
        SIMPLE  =                    T / Written by IDL:  Thu Jun  6 19:03:59 2002
        BITPIX  =                   16 /
        NAXIS   =                    2 /
        NAXIS1  =                 1024 /
        NAXIS2  =                 1024 /
        FILENAME= '32088304.fts'       /
        FILEORIG= '020521_001853.img'  /
        DATE    = '2002/06/06 23:03:55.204' /
        DATE-OBS= '2002/05/21'         /
        TIME-OBS= '00:18:06.516'       /
        P1COL   =                   20 /
        P1ROW   =                    1 /
        P2COL   =                 1043 /
        P2ROW   =                 1024 /
        VERSION =                    2 /
        EXPTIME =              19.0996 /
        EXP0    =              17.0000 /
        EXPCMD  =              17.0000 /
        EXP1    =              1.88135 /
        EXP2    =              3.17676 /
        EXP3    =              2.09961 /
        TELESCOP= 'SOHO    '           /
        INSTRUME= 'LASCO   '           /
        DETECTOR= 'C3      '           /
        READPORT= 'C       '           /
        SUMROW  =                    0 /
        SUMCOL  =                    0 /
        LEBXSUM =                    1 /
        LEBYSUM =                    1 /
        SHUTTR  =                    0 /
        LAMP    =                    0 /
        FILTER  = 'Clear   '           /
        POLAR   = 'Clear   '           /
        LP_NUM  = 'Normal  '           /
        OS_NUM  =                 3390 /
        IMGCTR  =                  717 /
        IMGSEQ  =                    0 /
        COMPRSSN= 'X#      '           /
        HCOMP_SF=                   32 /
        MID_DATE=                52415 /
        MID_TIME=              1096.07 /
        PLATESCL=              56.0000 /
        OFFSET  =              378.876 /
        IMAGE_CT=                  717 /
        SEQ_NUM =                    0 /
        OBT_TIME=        1.4006315E+09 /
        R1COL   =                   20 /
        R1ROW   =                    1 /
        R2COL   =                 1043 /
        R2ROW   =                 1024 /
        EFFPORT = 'C       '           /
        RECTIFY = 'TRUE    '           /
        DATAMIN =              376.000 /
        DATAMAX =              11183.0 /
        DATAZER =               101375 /
        DATASAT =                    0 /
        DATAAVG =              1886.68 /
        DATASIG =              1630.27 /
        DATAP01 =                  396 /
        DATAP10 =                  780 /
        DATAP25 =                  968 /
        DATAP75 =                 2029 /
        DATAP90 =                 4213 /
        DATAP95 =                 5959 /
        DATAP98 =                 7190 /
        DATAP99 =                 8023 /
        CRPIX1  =            517.95599 /
        CRPIX2  =            532.63202 /
        CRVAL1  =              0.00000 /
        CRVAL2  =              0.00000 /
        CROTA1  =              0.00000 /
        CROTA2  =              0.00000 /
        CTYPE1  = 'SOLAR-X '           /
        CTYPE2  = 'SOLAR-Y '           /
        CUNIT1  = 'ARCSEC  '           /
        CUNIT2  = 'ARCSEC  '           /
        CDELT1  =            56.000000 /
        CDELT2  =            56.000000 /
        HISTORY offset_bias.pro	1.24 12/13/01, 378.876
        HISTORY V15 18 Jun 1997 MAKE_FITS_HDR
        HISTORY 2
        """)
    header = fits.Header.fromstring(raw_header, sep='\n')
    data = np.random.rand(header['naxis1'], header['naxis2'])
    return Map(data, header)


def lasco():
    fitspath = glob.glob(os.path.join(path, "lasco_c2_25299383_s.fts"))
    return Map(fitspath)


@pytest.fixture(params=[lasco, lascoc3], ids=['C2', 'C3'])
def lasco_map(request):
    return request.param()


@pytest.fixture
def lasco_helioviewer():
    jp2path = glob.glob(os.path.join(
        path, "2013_05_13__16_54_06_137__SOHO_LASCO_C3_white-light.jp2"))
    return Map(jp2path)


def test_fitstoLASCO(lasco_map):
    """Tests the creation of LASCOMap using FITS."""
    assert isinstance(lasco_map, LASCOMap)


def test_is_datasource_for(lasco_map):
    """Test the is_datasource_for method of LASCOMap.
    Note that header data to be provided as an argument
    can be a MetaDict object."""
    assert lasco_map.is_datasource_for(lasco_map.data, lasco_map.meta)


def test_measurement(lasco_map):
    """Tests the measurement property of the LASCOMap object."""
    assert lasco_map.measurement == "white-light"


def test_wavelength(lasco_map):
    """Tests wavelength property."""
    assert lasco_map.wavelength is None


def test_date(lasco_map):
    assert lasco_map.date == parse_time(
        {'C2': '2009-02-28T00:05:33.380',
         'C3': '2002-05-21T00:18:06.516'}[lasco_map.detector])


def test_nickname(lasco_map):
    assert lasco_map.nickname == {'C2': 'LASCO-C2 Orange',
                                  'C3': 'LASCO-C3 Clear'}[lasco_map.detector]


def test_observatory(lasco_map):
    """Tests the observatory property of the LASCOMap object."""
    assert lasco_map.observatory == "SOHO"


def test_norm_clip(lasco_map):
    # Tests that the default normalizer has clipping disabled
    assert not lasco_map.plot_settings['norm'].clip


@skip_glymur
def test_helioviewer_rotation(lasco_map, lasco_helioviewer):
    """Tests that rotation metadata is correctly removed
    for JPEG2000 images provided by Helioviewer.org."""
    rmatrix = {'C2': [[0.999966, -0.008296], [0.008296, 0.999966]],
               'C3': [[1, 0], [0, 1]]}[lasco_map.detector]
    np.testing.assert_allclose(lasco_map.rotation_matrix, rmatrix, rtol=1e-6)
    np.testing.assert_array_equal(lasco_helioviewer.rotation_matrix, [[1., 0.], [0., 1.]])


def test_wcs(lasco_map):
    # Smoke test that WCS is valid and can transform from pixels to world coordinates
    with pytest.warns(SunpyMetadataWarning, match='Missing metadata for observer'):
        lasco_map.pixel_to_world(0*u.pix, 0*u.pix)



"""
=============================
Drawing and using a Great Arc
=============================

How to define and draw a great arc on an image of the
Sun, and to extract intensity values along that arc.
"""
import matplotlib.pyplot as plt
import numpy as np

import astropy.units as u
from astropy.coordinates import SkyCoord

import sunpy.map
from sunpy.coordinates.utils import GreatArc
from sunpy.data.sample import AIA_171_IMAGE

###############################################################################
# We start with the sample data.

m = sunpy.map.Map(AIA_171_IMAGE)

###############################################################################
# Let's define the start and end coordinates of the arc.

start = SkyCoord(735 * u.arcsec, -471 * u.arcsec, frame=m.coordinate_frame)
end = SkyCoord(-100 * u.arcsec, 800 * u.arcsec, frame=m.coordinate_frame)

###############################################################################
# Create the great arc between the start and end points.

great_arc = GreatArc(start, end)

###############################################################################
# Plot the great arc on the Sun.

fig = plt.figure()
ax = fig.add_subplot(projection=m)
m.plot(axes=ax, clip_interval=(1, 99.99)*u.percent)
ax.plot_coord(great_arc.coordinates(), color='c')

plt.show()

###############################################################################
# Now we can calculate the nearest integer pixels of the data that correspond
# to the location of arc.

pixels = np.asarray(np.rint(m.world_to_pixel(great_arc.coordinates())), dtype=int)
x = pixels[0, :]
y = pixels[1, :]

###############################################################################
# Get the intensity along the arc from the start to the end point.
intensity_along_arc = m.data[y, x]

###############################################################################
# Define the angular location of each pixel along the arc from the start point
# to the end.

angles = great_arc.inner_angles().to(u.deg)

###############################################################################
# Plot the intensity along the arc from the start to the end point.

plt.plot(angles, intensity_along_arc)
plt.xlabel(f'Distance along arc [{angles.unit}]')
plt.ylabel(f'Intensity [{m.unit}]')

plt.show()



"""
=======================================
Drawing a latitude-longitude quadrangle
=======================================

How to draw a latitude-longitude quadrangle on a map.
"""
import matplotlib.pyplot as plt

import astropy.units as u
from astropy.coordinates import SkyCoord

import sunpy.map
from sunpy.coordinates import HeliographicStonyhurst
from sunpy.data.sample import AIA_171_IMAGE

###############################################################################
# The purpose of this example is to demonstrate how to draw a quadrangle on a
# map using :meth:`~sunpy.map.GenericMap.draw_quadrangle`.  A quadrangle has
# edges aligned with lines of constant latitude and longitude in some
# coordinate system. We start with the sample AIA image.

aia = sunpy.map.Map(AIA_171_IMAGE)

###############################################################################
# Now let's define the bottom-left corner of the quadrangle.  Note that we
# define it in a different coordinate frame (heliographic Stonyhurst) than the
# coordinate frame of the map (which is helioprojective Cartesian).

bottom_left = SkyCoord(30*u.deg, -10*u.deg,
                       frame=HeliographicStonyhurst, obstime=aia.date)

###############################################################################
# Now let's draw a quadrangle on the map, with a width of 20 degrees and a
# height of 60 degrees.  Because the coordinate frame of ``bottom_left`` is
# in `~sunpy.coordinates.frames.HeliographicStonyhurst`, the width and height
# correspond to the longitude and latitude directions, respectively, in that
# coordinate frame.

fig = plt.figure()
ax = fig.add_subplot(projection=aia)
aia.plot(axes=ax, clip_interval=(1, 99.99)*u.percent)
aia.draw_grid(axes=ax)
aia.draw_quadrangle(bottom_left, axes=ax, width=20*u.deg, height=60*u.deg,
                    edgecolor='blue', linewidth=2)
ax.plot_coord(bottom_left, 'x', color='red')

plt.show()



"""
======================================
Searching and downloading from the VSO
======================================

How to download data from the VSO with Fido.
"""
import astropy.units as u

from sunpy.net import Fido
from sunpy.net import attrs as a

###############################################################################
# `sunpy.net.Fido` is the primary interface to search for and download data and
# will search the VSO when appropriate. The following example searches for all
# SOHO/EIT images between the times defined below by defining a
# timerange (`~sunpy.net.attrs.Time`) and the instrument (`~sunpy.net.attrs.Instrument`).

attrs_time = a.Time('2005/01/01 00:10', '2005/01/01 00:15')
result = Fido.search(attrs_time, a.Instrument.eit)

###############################################################################
# Let's inspect the results.

print(result)

###############################################################################
# The following shows how to download the results. If we
# don't provide a path it will download the file into the sunpy data directory.
# The output provides the path of the downloaded files.

downloaded_files = Fido.fetch(result)
print(downloaded_files)

###############################################################################
# More complicated queries can be constructed by using relational operators.
# For example, it is possible to query two wavelengths at the same time with
# the OR operator (|).

result = Fido.search(a.Time('2020/03/04 00:00', '2020/03/04 00:02'),
                     a.Instrument.aia,
                     a.Wavelength(171*u.angstrom) | a.Wavelength(94*u.angstrom))
print(result)

###############################################################################
# We can even combine entire queries in this manner.
# Here we will define two searches for the AIA and HMI data.
# But unlike other examples, we have to ``&`` the individual queries.

search_aia = (a.Time('2020/03/04 00:00', '2020/03/04 00:01') & a.Instrument.aia)
search_hmi = (a.Time('2020/03/04 00:00', '2020/03/04 00:01')
              & a.Instrument.hmi & a.Physobs.los_magnetic_field)

result = Fido.search(search_aia | search_hmi)
print(result)



"""
======================================
Downloading and plotting LASCO C2 data
======================================

How to download SOHO/LASCO C2 data with Fido and make a plot.
"""
import matplotlib.pyplot as plt

import sunpy.map
from sunpy.net import Fido
from sunpy.net import attrs as a

###############################################################################
# In order to download the required data, we use `sunpy.net.Fido`, a downloader client.
# We define two search variables: a timerange and the instrument.

timerange = a.Time('2002/05/24 11:06', '2002/05/24 11:07')
instrument = a.Instrument.lasco
detector = a.Detector.c2
result = Fido.search(timerange, instrument, detector)

###############################################################################
# Let's inspect the result.
print(result)

###############################################################################
# The following shows how to download the results. If we
# don't provide a path it will download the file into the sunpy data directory.
# The output provides the path of the downloaded files.

downloaded_files = Fido.fetch(result)
print(downloaded_files)

###############################################################################
# We can then load a downloaded file into a sunpy map and plot it.

lascomap = sunpy.map.Map(downloaded_files[0])
plt.figure()
lascomap.plot()
plt.show()



"""
This module provides a translation layer between the HEK and the VSO.
It allows you to acquire records of data that are available via the
VSO, based on the data in HEK event entries.

.. warning::
    This module is in beta and maybe unstable.

"""
from sunpy.net.hek2vso.hek2vso import *



"""
=====================================================
Reprojecting to a Map Projection with a Custom Origin
=====================================================

In this example, we show how to reproject a map to a map projection with a
custom origin. Here, we choose the target map projection to be the
`azimuthal equidistant projection <https://en.wikipedia.org/wiki/Azimuthal_equidistant_projection>`__,
also known as the Postel projection, which has useful properties relative to a
specified origin of the projection. If a different map projection is desired,
modifying this example is straightforward.

"""

import matplotlib.pyplot as plt

import astropy.units as u
from astropy.coordinates import SkyCoord

import sunpy.map
from sunpy.data.sample import AIA_171_IMAGE

###############################################################################
# We will use one of the AIA images from the sample data. We fix the range of
# values for the Map's normalizer for a prettier image.

aia_map = sunpy.map.Map(AIA_171_IMAGE)
aia_map.plot_settings['norm'].vmin = 0
aia_map.plot_settings['norm'].vmax = 10000

###############################################################################
# Next, we create a `~astropy.coordinates.SkyCoord` to define the custom origin
# of the map projection. Here, we are going to center the projection at the
# helioprojective coordinates of a particular active region. We want our map
# projection to be in heliographic Stonyhurst coordinates, so we transform the
# origin coordinate accordingly.

origin_hpc = SkyCoord(735*u.arcsec, -340*u.arcsec, frame=aia_map.coordinate_frame)
origin = origin_hpc.heliographic_stonyhurst

###############################################################################
# We then create a FITS-WCS header that includes our custom origin coordinate.
# The azimuthal equidistant projection is specified by the code ``"ARC"``.
# See :doc:`astropy:wcs/supported_projections` for the projection codes for
# other projections.

out_shape = (750, 750)
out_header = sunpy.map.make_fitswcs_header(
    out_shape,
    origin,
    scale=[0.4, 0.4]*u.deg/u.pix,
    projection_code="ARC"
)

###############################################################################
# We reproject the map to our FITS-WCS header and copy over the plot settings.

out_map = aia_map.reproject_to(out_header)
out_map.plot_settings = aia_map.plot_settings

###############################################################################
# Finally, we plot both the original and reprojected maps side by side.

fig = plt.figure(figsize=(8, 4))

# sphinx_gallery_defer_figures

###############################################################################
# Plot the original AIA map, with the active region circled in red and the
# heliographic grid and solar limb in blue.

ax = fig.add_subplot(1, 2, 1, projection=aia_map)
aia_map.plot(axes=ax)
aia_map.draw_grid(axes=ax, color='blue')
aia_map.draw_limb(axes=ax, color='blue')
ax.plot_coord(origin, 'o', color='red', fillstyle='none', markersize=20)

# sphinx_gallery_defer_figures

###############################################################################
# Plot the reprojected AIA map, again with the active region circled in red and
# the heliographic grid and solar limb in blue.

ax = fig.add_subplot(1, 2, 2, projection=out_map)
out_map.plot(axes=ax)
out_map.draw_grid(axes=ax, color='blue')
out_map.draw_limb(axes=ax, color='blue')
ax.plot_coord(origin, 'o', color='red', fillstyle='none', markersize=20)
ax.set_title('Postel projection centered at ROI', y=-0.1)
plt.show()



from sunpy.util.exceptions import warn_deprecated
from . import _fits
from ._fits import *  # NoQA

__doc__ = _fits.__doc__
__all__ = _fits.__all__

warn_deprecated("The `sunpy.io.fits` module is deprecated, as it was designed "
                "for internal use. Use the `astropy.fits.io` module instead "
                "for more generic functionality to read FITS files.")



"""A Composite Map class

Author: `Keith Hughitt <keith.hughitt@nasa.gov>`
"""
import matplotlib.pyplot as plt
from matplotlib.collections import Collection, QuadMesh
from matplotlib.contour import ContourSet, QuadContourSet
from matplotlib.image import AxesImage, _ImageBase

import astropy.units as u

from sunpy.map import GenericMap
from sunpy.util import expand_list, get_keywords, get_set_methods
from sunpy.util.decorators import add_common_docstring
from sunpy.visualization import axis_labels_from_ctype, peek_show, wcsaxes_compat

__all__ = ['CompositeMap']

__author__ = "Keith Hughitt"
__email__ = "keith.hughitt@nasa.gov"

# Valid keyword arguments for each plotting method
ACCEPTED_IMSHOW_KWARGS = get_keywords(
    [GenericMap.plot, plt.Axes.imshow, AxesImage.__init__, _ImageBase.__init__]
) | get_set_methods(AxesImage)

ACCEPTED_PCOLORMESH_KWARGS = (get_keywords(
    [GenericMap.plot, plt.Axes.pcolormesh, QuadMesh.__init__, Collection.__init__]
) | get_set_methods(QuadMesh)) - {
    'color', 'ec', 'edgecolor', 'facecolor', 'linestyle', 'linestyles',
    'linewidth', 'linewidths', 'ls', 'lw'
}

ACCEPTED_CONTOUR_KWARGS = get_keywords(
    [GenericMap.draw_contours, ContourSet.__init__, QuadContourSet._process_args]
)


class CompositeMap:
    """
    CompositeMap(map1 [,map2,..])

    A Composite Map class

    Parameters
    ----------
    args : [`~sunpy.map.Map` | string]
        One or more map of filepaths

    Examples
    --------
    >>> import sunpy.map
    >>> import sunpy.data.sample  # doctest: +REMOTE_DATA
    >>> comp_map = sunpy.map.Map(sunpy.data.sample.AIA_171_IMAGE,
    ...                          sunpy.data.sample.EIT_195_IMAGE,
    ...                          composite=True)  # doctest: +REMOTE_DATA
    >>> comp_map.add_map(sunpy.map.Map(sunpy.data.sample.RHESSI_IMAGE))  # doctest: +REMOTE_DATA
    >>> comp_map.peek()  # doctest: +SKIP

    """

    def __init__(self, *args, **kwargs):
        self._maps = expand_list(args)

        for m in self._maps:
            if not isinstance(m, GenericMap):
                raise ValueError(
                    'CompositeMap expects pre-constructed map objects.')

        # Default alpha and zorder values
        alphas = [1] * len(self._maps)
        zorders = list(range(0, 10 * len(self._maps), 10))
        levels = [False] * len(self._maps)

        # Set z-order and alpha values for the map
        for i, m in enumerate(self._maps):
            m.zorder = zorders[i]
            m.alpha = alphas[i]
            m.levels = levels[i]

    def add_map(self, amap, zorder=None, alpha=1, levels=False):
        """Adds a map to the CompositeMap.

        Parameters
        ----------
        amap : `~sunpy.map.GenericMap` or subclass
            Map instance to be added
        zorder : `int`
            The index to use when determining where the map should lie along
            the z-axis; maps with higher z-orders appear above maps with lower
            z-orders.
        alpha : `float`
            Opacity at which the map should be displayed. An alpha value of 0
            results in a fully transparent image while an alpha value of 1
            results in a fully opaque image. Values between result in semi-
            transparent images.
        """
        if zorder is None:
            zorder = max([m.zorder for m in self._maps]) + 10

        amap.zorder = zorder
        amap.alpha = alpha
        amap.levels = levels

        self._maps.append(amap)

    def remove_map(self, index):
        """Removes and returns the map with the given index.

        Parameters
        ----------
        index : `int`
            The index of the map in the composite map.

        Returns
        -------
        `sunpy.map.CompositeMap`
            A composite map with the map indexed by 'index' removed from the
            composite map.
        """
        return self._maps.pop(index)

    def list_maps(self):
        """Prints a list of the currently included maps."""
        print([m.__class__ for m in self._maps])

    def get_map(self, index):
        """Returns the map with given index """
        return self._maps[index]

    def get_alpha(self, index=None):
        """
        Returns the alpha-channel value for a layer in the composite image.
        """
        if index is None:
            return [_map.alpha for _map in self._maps]
        else:
            return self._maps[index].alpha

    def get_levels(self, index=None):
        """Returns the list of contour levels for a map within the
        composite.

        Parameters
        ----------
        index : {`int` | None}
            The index of the map in the composite map.

        Returns
        -------
        `list`
            A list of the contour levels of map at index 'index' in the
            composite map.  If index is None, then the contour levels of all
            the maps are returned as a list of lists.
        """
        if index is None:
            return [_map.levels for _map in self._maps]
        else:
            return self._maps[index].levels

    def get_plot_settings(self, index=None):
        """Returns the plot settings for a map within the composite map.

        Parameters
        ----------
        index : {`int` | None}
            The index of the map in the composite map.

        Returns
        -------
        {`dict` | `list`}
            The plot settings of the map(s) in the composite map.  If None
            then the plot settings of all the maps are returned in a list.
        """

        if index is None:
            return [_map.plot_settings for _map in self._maps]
        else:
            return self._maps[index].plot_settings

    def get_zorder(self, index=None):
        """Returns the layering preference (z-order) for a map within the
        composite.

        Parameters
        ----------
        index : {`int` | None}
            The index of the map in the composite map.

        Returns
        -------
        {`float` | `list`}
            The layering order (z-order) of the map(s) in the composite
            map.  If None then the layering order of all the maps is returned in
            a list.
        """
        if index is None:
            return [_map.zorder for _map in self._maps]
        else:
            return self._maps[index].zorder

    def set_alpha(self, index, alpha):
        """Sets the alpha-channel value for a layer in the composite image.

        Parameters
        ----------
        index : `int`
            The index of the map in the composite map.
        alpha : `float`
            A float in the range 0 to 1.  Increasing values of alpha decrease
            the transparency of the layer (0 is complete transparency, 1
            indicates the layer will be completely opaque).

        Returns
        -------
        `~sunpy.map.CompositeMap`
            A composite map with alpha-channel value 'alpha' at layer 'index'.
        """
        if 0 <= alpha <= 1:
            self._maps[index].alpha = alpha
        else:
            raise OutOfRangeAlphaValue("Alpha value must be between 0 and 1.")

    def set_levels(self, index, levels, percent=False):
        """
        Sets the contour levels for a layer in the composite image.

        Parameters
        ----------
        index : `int`
            The index of the map in the composite map.
        levels : array-like
            The contour levels.
        percent : `bool`
            If True, the input 'levels' are interpreted as percentages relative
            to the maximum value of the data in layer 'index' of the composite
            map.  If False, the contour levels are set directly from 'levels'.

        Returns
        -------
        `~sunpy.map.CompositeMap`
            A composite map with contour levels 'levels' at layer 'index'.
        """
        if percent is False:
            self._maps[index].levels = levels
        else:
            self._maps[index].levels = u.Quantity(levels, u.percent)

    def set_plot_settings(self, index, plot_settings):
        """Sets the plot settings for a layer in the composite image.

        Parameters
        ----------
        index : `int`
            The index of the map in the composite map.
        plot_settings : `dict`
            A dictionary of the form

        Returns
        -------
        `~sunpy.map.CompositeMap`
            A composite map with plot settings 'plot_settings' at layer
            'index'.
        """
        self._maps[index].plot_settings = plot_settings

    def set_zorder(self, index, zorder):
        """Set the layering order (z-order) for a map within the
        composite.

        Parameters
        ----------
        index : `int`
            The index of the map in the composite map.
        zorder : `int`
            The layer order.

        Returns
        -------
        `~sunpy.map.CompositeMap`
            A composite map with the map at layer 'index' having layering order
            'zorder'.
        """
        self._maps[index].zorder = zorder

    def draw_limb(self, index=None, axes=None, **kwargs):
        """Draws a circle representing the solar limb.

        Parameters
        ----------
        index : `int`
            Map index to use to plot limb.
        axes : `matplotlib.axes.Axes` or None
            Axes to plot limb on or None to use current axes.

        Returns
        -------
        `matplotlib.axes.Axes`

        Notes
        -----
        Keyword arguments are passed onto `sunpy.map.mapbase.GenericMap.draw_limb`.
        """
        if index is None:
            for i, amap in enumerate(self._maps):
                if hasattr(amap, 'rsun_obs'):
                    index = i
                    break

        index_check = hasattr(self._maps[index], 'rsun_obs')
        if not index_check or index is None:
            raise ValueError("Specified index does not have all"
                             " the required attributes to draw limb.")

        return self._maps[index].draw_limb(axes=axes, **kwargs)

    @u.quantity_input
    def draw_grid(self, index=None, axes=None, grid_spacing: u.deg = 20*u.deg, **kwargs):
        """Draws a grid over the surface of the Sun.

        Parameters
        ----------
        index : int
            Index to determine which map to use to draw grid.
        axes : `~matplotlib.axes.Axes` or None
            Axes to plot limb on or None to use current axes.
        grid_spacing : `float`
            Spacing (in degrees) for longitude and latitude grid.

        Returns
        -------
        `matplotlib.axes.Axes` object

        Notes
        -----
        Keyword arguments are passed onto `sunpy.map.mapbase.GenericMap.draw_grid`.
        """
        needed_attrs = ['rsun_meters', 'dsun', 'heliographic_latitude',
                        'heliographic_longitude']
        if index is None:
            for i, amap in enumerate(self._maps):
                if all([hasattr(amap, k) for k in needed_attrs]):
                    index = i
                    break

        index_check = all([hasattr(self._maps[index], k) for k in needed_attrs])
        if not index_check or index is None:
            raise ValueError("Specified index does not have all"
                             " the required attributes to draw grid.")

        ax = self._maps[index].draw_grid(axes=axes, grid_spacing=grid_spacing, **kwargs)
        return ax

    @add_common_docstring(
        ACCEPTED_IMSHOW_KWARGS=sorted(ACCEPTED_IMSHOW_KWARGS),
        ACCEPTED_PCOLORMESH_KWARGS=sorted(ACCEPTED_PCOLORMESH_KWARGS),
        ACCEPTED_CONTOUR_KWARGS=sorted(ACCEPTED_CONTOUR_KWARGS)
    )
    def plot(self, axes=None, annotate=True,
             title="SunPy Composite Plot", **matplot_args):
        """Plots the composite map object by calling :meth:`~sunpy.map.GenericMap.plot`
        or :meth:`~sunpy.map.GenericMap.draw_contours`.

        By default, each map is plotted as an image.  If a given map has levels
        defined (via :meth:`~sunpy.map.CompositeMap.set_levels`), that map will instead
        be plotted as contours.

        Parameters
        ----------
        axes : `~matplotlib.axes.Axes` or None
            If provided the image will be plotted on the given axes. Else the
            current matplotlib axes will be used.
        annotate : `bool`
            If true, the data is plotted at it's natural scale; with
            title and axis labels.
        title : `str`
            Title of the composite map.
        **matplot_args : `dict`
            Any additional Matplotlib arguments that should be used
            when plotting.

        Returns
        -------
        ret : `list`
            List of axes image or quad contour sets that have been plotted.

        Notes
        -----
        Images are plotted using either `~matplotlib.axes.Axes.imshow` or
        `~matplotlib.axes.Axes.pcolormesh`, and contours are plotted using
        `~matplotlib.axes.Axes.contour`.
        The Matplotlib arguments accepted by the plotting method are passed to it.
        (For compatability reasons, we enforce a more restrictive set of
        accepted `~matplotlib.axes.Axes.pcolormesh` arguments.)
        If any Matplotlib arguments are not used by any plotting method,
        a ``TypeError`` will be raised.
        The ``sunpy.map.compositemap`` module includes variables which list the
        full set of arguments passed to each plotting method. These are:

        >>> import sunpy.map.compositemap
        >>> sorted(sunpy.map.compositemap.ACCEPTED_IMSHOW_KWARGS)
        {ACCEPTED_IMSHOW_KWARGS}
        >>> sorted(sunpy.map.compositemap.ACCEPTED_PCOLORMESH_KWARGS)
        {ACCEPTED_PCOLORMESH_KWARGS}
        >>> sorted(sunpy.map.compositemap.ACCEPTED_CONTOUR_KWARGS)
        {ACCEPTED_CONTOUR_KWARGS}

        If a transformation is required to overlay the maps with the correct
        alignment, the plot limits may need to be manually set because
        Matplotlib autoscaling may not work as intended.
        """

        # If axes are not provided, create a WCSAxes based on the first map
        if not axes:
            axes = wcsaxes_compat.gca_wcs(self._maps[0].wcs)

        if annotate:
            axes.set_xlabel(axis_labels_from_ctype(self._maps[0].coordinate_system[0],
                                                   self._maps[0].spatial_units[0]))
            axes.set_ylabel(axis_labels_from_ctype(self._maps[0].coordinate_system[1],
                                                   self._maps[0].spatial_units[1]))
            axes.set_title(title)

        # Checklist to determine unused keywords in `matplot_args`
        unused_kwargs = set(matplot_args.keys())

        # Define a list of plotted objects
        ret = []
        # Plot layers of composite map
        for m in self._maps:
            # Parameters for plotting
            params = {
                "alpha": m.alpha,
                "zorder": m.zorder,
            }
            params.update(matplot_args)

            # The request to show a map layer rendered as a contour is indicated by a
            # non False levels property.
            if m.levels is False:
                # We tell GenericMap.plot() that we need to autoalign the map
                if wcsaxes_compat.is_wcsaxes(axes):
                    params['autoalign'] = True

                # Filter `matplot_args`
                if params.get('autoalign', None) in (True, 'pcolormesh'):
                    accepted_kwargs = ACCEPTED_PCOLORMESH_KWARGS
                else:
                    accepted_kwargs = ACCEPTED_IMSHOW_KWARGS
                for item in matplot_args.keys():
                    if item not in accepted_kwargs:
                        del params[item]
                    else:  # mark as used
                        unused_kwargs -= {item}

                params['annotate'] = False
                ret.append(m.plot(**params))
            else:
                # Filter `matplot_args`
                for item in matplot_args.keys():
                    if item not in ACCEPTED_CONTOUR_KWARGS:
                        del params[item]
                    else:  # mark as used
                        unused_kwargs -= {item}

                ret.append(m.draw_contours(m.levels, **params))

                # Set the label of the first line so a legend can be created
                ret[-1].collections[0].set_label(m.name)

        if len(unused_kwargs) > 0:
            raise TypeError(f'plot() got unexpected keyword arguments {unused_kwargs}')

        # Adjust axes extents to include all data
        axes.axis('image')

        # Set current image (makes colorbar work)
        plt.sci(ret[0])
        return ret

    @peek_show
    def peek(self, colorbar=True, draw_limb=True, draw_grid=False, **matplot_args):
        """
        Displays a graphical overview of the data in this object for user evaluation.
        For the creation of plots, users should instead use the `~sunpy.map.CompositeMap.plot`
        method and Matplotlib's pyplot framework.

        Parameters
        ----------
        colorbar : `bool` or `int`
            Whether to display a colorbar next to the plot.
            If specified as an integer a colorbar is plotted for that index.
        draw_limb : `bool`
            If true, draws a circle representing the solar limb.
        draw_grid : `bool`
            If true, draws a grid over the surface of the Sun.
        **matplot_args : dict
            Matplotlib Any additional imshow arguments that should be used
            when plotting.
        """

        # Create a figure and add title and axes
        figure = plt.figure()

        axes = figure.add_subplot(111, projection=self._maps[0])

        ret = self.plot(axes=axes, **matplot_args)

        if not isinstance(colorbar, bool) and isinstance(colorbar, int):
            figure.colorbar(ret[colorbar])
        elif colorbar:
            plt.colorbar()
        if draw_limb:
            self.draw_limb(axes=axes)

        if isinstance(draw_grid, bool):
            if draw_grid:
                self.draw_grid(axes=axes)

        elif isinstance(draw_grid, (int, float)):
            self.draw_grid(axes=axes, grid_spacing=draw_grid)
        else:
            raise TypeError("draw_grid should be bool, int, long or float")

        return figure


class OutOfRangeAlphaValue(ValueError):
    """Exception to raise when an alpha value outside of the range 0-1 is
    requested.
    """



import datetime
from unittest import mock

import pytest

from sunpy.net import Fido
from sunpy.net import attrs as a
from sunpy.net._attrs import Instrument, Time
from sunpy.net.dataretriever.client import QueryResponse
from sunpy.net.dataretriever.sources import noaa
from sunpy.tests.helpers import no_vso
from sunpy.time import parse_time


@pytest.fixture
def indices_client():
    return noaa.NOAAIndicesClient()


@pytest.fixture
def predict_client():
    return noaa.NOAAPredictClient()


@pytest.fixture
def srs_client():
    return noaa.SRSClient()


def mock_query_object(start_date, end_date):
    """
    Creation of a QueryResponse object, and prefill some
    downloaded data from noaa.NOAAIndicesClient().fetch(Time('20 ..)
    """
    # Create a mock Query Response object
    start = parse_time(start_date)
    end = parse_time(end_date)
    obj = {
        'Start Time': parse_time(start),
        'End Time': parse_time(end),
        'Instrument': 'NOAA-Indices',
        'Physobs': 'sunspot number',
        'Source': 'SIDC',
        'Provider': 'SWPC',
        'url': 'https://services.swpc.noaa.gov/json/solar-cycle/observed-solar-cycle-indices.json'
    }
    results = QueryResponse([obj], client=noaa.NOAAIndicesClient())
    return results


@pytest.mark.remote_data
def test_fetch_working(indices_client, tmpdir):
    """
    Tests if the online server for noaa is working.
    Uses the url : https://services.swpc.noaa.gov/json/solar-cycle/observed-solar-cycle-indices.json
    """
    qr1 = indices_client.search(Time('2012/10/4', '2012/10/6'),
                                Instrument('noaa-indices'))
    # Mock QueryResponse object
    mock_qr = mock_query_object('2012/10/4', '2012/10/6')
    # Compare if two objects have the same attribute
    mock_qr = mock_qr[0]
    qr = qr1[0]
    assert mock_qr['Source'] == qr['Source']
    assert mock_qr['Provider'] == qr['Provider']
    assert mock_qr['Physobs'] == qr['Physobs']
    assert mock_qr['Instrument'] == qr['Instrument']
    assert mock_qr['url'] == qr['url']
    target_dir = tmpdir.mkdir("down")
    download_list = indices_client.fetch(qr1, path=target_dir)
    assert len(download_list) == len(qr1)
    assert download_list[0].split('/')[-1] == 'observed-solar-cycle-indices.json'


@pytest.mark.parametrize(
    "timerange,url_start,url_end",
    [(Time('1995/06/03', '1995/06/04'),
      'https://services.swpc.noaa.gov/json/solar-cycle/observed-solar-cycle-indices.json',
      'https://services.swpc.noaa.gov/json/solar-cycle/observed-solar-cycle-indices.json'),
     (Time('2008/06/01', '2008/06/02'),
      'https://services.swpc.noaa.gov/json/solar-cycle/observed-solar-cycle-indices.json',
      'https://services.swpc.noaa.gov/json/solar-cycle/observed-solar-cycle-indices.json')])
def test_get_url_for_time_range(indices_client, timerange, url_start, url_end):
    resp = indices_client.search(timerange)
    urls = [i['url'] for i in resp]
    assert isinstance(urls, list)
    assert urls[0] == url_start
    assert urls[-1] == url_end


def test_can_handle_query():
    ans1 = noaa.NOAAIndicesClient._can_handle_query(
        Time('2012/8/9', '2012/8/10'), Instrument('noaa-indices'))
    assert ans1
    ans2 = noaa.NOAAIndicesClient._can_handle_query(
        Time('2012/7/7', '2012/7/7'))
    assert not ans2
    ans3 = noaa.NOAAIndicesClient._can_handle_query(
        Time('2012/8/9', '2012/8/10'), Instrument('eve'))
    assert not ans3


@mock.patch('sunpy.net.dataretriever.sources.noaa.NOAAIndicesClient.search',
            return_value=mock_query_object('2012/8/9', '2012/8/10'))
def test_query(mock_search, indices_client):
    qr1 = indices_client.search(
        Time('2012/8/9', '2012/8/10'), Instrument('noaa-indices'))
    assert isinstance(qr1, QueryResponse)
    assert len(qr1) == 1
    assert qr1.time_range().start == parse_time('2012/08/09')
    assert qr1.time_range().end == parse_time('2012/08/10')


@mock.patch('sunpy.net.dataretriever.sources.noaa.NOAAIndicesClient.search',
            return_value=mock_query_object('2012/10/4', '2012/10/6'))
# The return value of download is irrelevant
@mock.patch('parfive.Downloader.download',
            return_value=None)
@mock.patch('parfive.Downloader.enqueue_file')
def test_fetch(mock_wait, mock_search, mock_enqueue, tmp_path, indices_client):
    path = tmp_path / "sub"
    path.mkdir()
    qr1 = indices_client.search(Time('2012/10/4', '2012/10/6'),
                                Instrument('noaa-indices'))
    indices_client.fetch(qr1, path=path / "{file}")

    # Here we assert that the `fetch` function has called the parfive
    # Downloader.enqueue_file method with the correct arguments. Everything
    # that happens after this point should either be tested in the
    # GenericClient tests or in parfive itself.
    assert mock_enqueue.called_once_with(("https://services.swpc.noaa.gov/json/solar-cycle/observed-solar-cycle-indices.json",
                                          path / "observed-solar-cycle-indices.json"))


@mock.patch('sunpy.net.dataretriever.sources.noaa.NOAAIndicesClient.search',
            return_value=mock_query_object('2012/10/4', '2012/10/6'))
# The return value of download is irrelevant
@mock.patch('parfive.Downloader.download',
            return_value=None)
@mock.patch('parfive.Downloader.enqueue_file')
@no_vso
def test_fido(mock_wait, mock_search, mock_enqueue, tmp_path, indices_client):
    path = tmp_path / "sub"
    path.mkdir()
    qr1 = Fido.search(Time('2012/10/4', '2012/10/6'),
                      Instrument('noaa-indices'))
    Fido.fetch(qr1, path=path)

    # Here we assert that the `fetch` function has called the parfive
    # Downloader.enqueue_file method with the correct arguments. Everything
    # that happens after this point should either be tested in the
    # GenericClient tests or in parfive itself.
    assert mock_enqueue.called_once_with(("https://services.swpc.noaa.gov/json/solar-cycle/observed-solar-cycle-indices.json",
                                          path / "observed-solar-cycle-indices.json"))


@no_vso
@pytest.mark.remote_data
def test_srs_unpack():
    qr = Fido.search(a.Instrument("soon") & a.Time("2015/01/01", "2015/01/01T23:59:29"))
    res = Fido.fetch(qr)
    assert len(res) == 1
    assert res.data[0].endswith("20150101SRS.txt")


@no_vso
@pytest.mark.remote_data
def test_srs_midyear():
    qr = Fido.search(a.Instrument("soon") & a.Time("2011/06/07", "2011/06/08T23:59:29"))
    res = sorted(Fido.fetch(qr))
    assert len(res) == 2
    assert res[0].endswith("20110607SRS.txt")
    assert res[-1].endswith("20110608SRS.txt")


@no_vso
@pytest.mark.remote_data
def test_srs_current_year():
    # Current year is nothing but text files, all older years should be tar files.
    year = datetime.date.today().year
    qr = Fido.search(a.Instrument("soon") & a.Time(f"{year}/01/01", f"{year}/01/01T23:59:29"))
    res = Fido.fetch(qr)
    assert len(res) <= 1
    if len(res):
        assert res.data[0].endswith(f"{year}0101SRS.txt")


@no_vso
@pytest.mark.remote_data
def test_srs_save_path(tmpdir):
    qr = Fido.search(a.Instrument.srs_table, a.Time("2016/10/01", "2016/10/02"))
    files = sorted(Fido.fetch(qr, path=str(tmpdir)))
    assert len(files) == 2
    assert files[0].endswith("20161001SRS.txt")
    assert files[1].endswith("20161002SRS.txt")


@pytest.mark.remote_data
@pytest.mark.filterwarnings('ignore:ERFA function')
def test_srs_out_of_range(srs_client):
    res = srs_client.search(a.Time('1995/01/01', '1995/02/01'))
    assert len(res) == 0
    res = srs_client.search(a.Time('2995/01/01', '2995/02/01'))
    assert len(res) == 0


@pytest.mark.remote_data
@pytest.mark.filterwarnings('ignore:ERFA function')
def test_srs_start_or_end_out_of_range(srs_client):
    res = srs_client.search(a.Time('1995/12/30', '1996/01/02'))
    assert len(res) == 1
    cur_year = datetime.date.today().year
    res = srs_client.search(a.Time(f'{cur_year}/01/01', f'{cur_year+2}/01/01'))
    assert len(res) > 0


@pytest.mark.remote_data
def test_tar_file_broken():
    # 2010 extracts out to 2010_SRS while other years do SRS only.
    results = Fido.search(a.Time("2010/5/1", "2010/5/2"), a.Instrument.soon)
    results = Fido.fetch(results)
    assert len(results) == 2


def test_no_time(predict_client, indices_client):
    res = indices_client.search(a.Instrument.noaa_indices)
    assert len(res) == 1
    res = predict_client.search(a.Instrument.noaa_predict)
    assert len(res) == 1


def test_attr_reg():
    assert a.Instrument.noaa_indices == a.Instrument("NOAA-Indices")
    assert a.Instrument.noaa_predict == a.Instrument("NOAA-Predict")
    assert a.Instrument.srs_table == a.Instrument("SRS-Table")
    assert a.Instrument.soon == a.Instrument("SOON")


def test_client_repr(indices_client):
    """
    Repr check
    """
    output = str(indices_client)
    assert output[:50] == 'sunpy.net.dataretriever.sources.noaa.NOAAIndicesCl'


def test_show():
    mock_qr = mock_query_object('2012/10/4', '2012/10/6')
    qrshow0 = mock_qr.show()
    qrshow1 = mock_qr.show('Source', 'Instrument')
    allcols = {'Start Time', 'End Time', 'Instrument', 'Physobs', 'Source', 'Provider', 'url'}
    assert not allcols.difference(qrshow0.colnames)
    assert qrshow1.colnames == ['Source', 'Instrument']
    assert qrshow0['Instrument'][0] == 'NOAA-Indices'



"""
This file will generate an asdf file in the test data, using the newest schema version.
"""
import asdf
import astropy.units as u

from sunpy.data.test import rootdir


def generate_asdf_tree(obj, filename):
    tree = {"object": obj}
    with asdf.AsdfFile(tree) as af:
        # TODO: Automatically determine filename based on tag used
        af.write_to(rootdir / filename)


if __name__ == "__main__":
    import sunpy.map
    test_map = rootdir / "aia_171_level1.fits"
    obj = sunpy.map.Map(test_map)
    obj = obj.resample((2, 2)*u.pix)
    #obj = obj.shift(10*u.arcsec, 0*u.arcsec)
    generate_asdf_tree(obj, "aiamap_genericmap_1.0.0.asdf")
